#    dep3.py -- DEP-3 compatible patch formatting
#    Copyright (C) 2011 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""DEP-3 style patch formatting."""

from bzrlib import diff

import time


def write_dep3_bug_line(f, bug_url, status):
    """Write a DEP-3 compatible line with a bug link.

    :param f: File-like object to write to
    :param bug_url: Bug URL
    :param status: Bug status (e.g. "fixed")
    """
    # For the moment, we only care about fixed bugs
    if status != "fixed":
        return
    if bug_url.startswith("http://bugs.debian.org/"):
        f.write("Bug-Debian: %s\n" % bug_url)
    else:
        # FIXME: Filter out Ubuntu bugs on Launchpad
        f.write("Bug: %s\n" % bug_url)


def write_dep3_patch_header(f, description=None, origin=None, forwarded=None,
        bugs=None, authors=None, revision_id=None, last_update=None,
        applied_upstream=None):
    """Write a DEP3 patch header.

    :param f: File-like object to write to
    :param description: Description of the patch
    :param origin: Single line describing the origin of the patch
    :param forwarded: Single line describing whether and how the patch was
        forwarded
    :param bugs: Set of bugs fixed in this patch
    :param authors: Authors of the patch
    :param revision_id: Relevant bzr revision id
    :param last_update: Last update timestamp
    :param applied_upstream: If the patch is applied upstream,
        an informal string describing where it was merged
    """
    # FIXME: Handle line breaks, etc sensibly
    if description is not None:
        description = description.strip("\n")
        description = description.replace("\n\n", "\n.\n")
        description = description.replace("\n", "\n ")
        f.write("Description: %s\n" % description)
    if origin is not None:
        f.write("Origin: %s\n" % origin)
    if forwarded is not None:
        f.write("Forwarded: %s\n" % forwarded)
    if authors is not None:
        for author in authors:
            f.write("Author: %s\n" % author)
    if bugs is not None:
        for bug_url, status in bugs:
            write_dep3_bug_line(f, bug_url, status)
    if last_update is not None:
        f.write("Last-Update: %s\n" % time.strftime("%Y-%m-%d",
            time.gmtime(last_update)))
    if applied_upstream is not None:
        f.write("Applied-Upstream: %s\n" % applied_upstream)
    if revision_id is not None:
        f.write("X-Bzr-Revision-Id: %s\n" % revision_id)
    f.write("\n")


def gather_bugs_and_authors(repository, interesting_revision_ids):
    """Gather bug and author information from revisions.

    :param interesting_revision_ids: Iterable of revision ids to check
    :return: Tuple of bugs, authors and highest found commit timestamp
    """
    authors = set()
    bugs = set()
    last_update = None
    for rev in repository.get_revisions(interesting_revision_ids):
        last_update = max(rev.timestamp, last_update)
        authors.update(rev.get_apparent_authors())
        bugs.update(rev.iter_bugs())
    return (bugs, authors, last_update)


def determine_applied_upstream(upstream_branch, feature_branch,
        feature_revid=None):
    """Check if a particular revision has been merged upstream.

    :param upstream_branch: Upstream branch object
    :param feature_branch: Feature branch
    :param feature_revid: Revision id in feature branch to check,
        defaults to feature_branch tip.
    :return: String that can be used for Applied-Upstream field
    """
    if feature_revid is None:
        feature_revid = feature_branch.last_revision()
    upstream_graph = feature_branch.repository.get_graph(
        upstream_branch.repository)
    merger = upstream_graph.find_lefthand_merger(feature_revid,
        upstream_branch.last_revision())
    if merger is not None:
        return "merged in revision %s" % (
            ".".join(str(x) for x in upstream_branch.revision_id_to_dotted_revno(merger)), )
    else:
        return "no"


def determine_forwarded(upstream_branch, feature_branch, feature_revid):
    """See if a branch has been forwarded to upstream.

    :param upstream_branch: Upstream branch object
    :param feature_branch: Feature branch
    :param feature_revid: Revision id in feature branch to check
    :return: String that can be used for Applied-Upstream field
    """
    # FIXME: Check for Launchpad merge proposals from feature_branch (or its
    # public_branch) to upstream_branch

    # Are there any other ways to see that a patch has been forwarded upstream?
    return None


def describe_origin(branch, revid):
    """Describe a tree for use in the origin field.

    :param branch: Branch to retrieve the revision from
    :param revid: Revision id
    """
    public_branch_url = branch.get_public_branch()
    if public_branch_url is not None:
        return "commit, %s, revision: %s" % (
            public_branch_url,
            ".".join(str(x) for x in branch.revision_id_to_dotted_revno(revid)), )
    else:
        return "commit, revision id: %s" % revid


def write_dep3_patch(f, branch, base_revid, target_revid, description=None,
        origin=None, forwarded=None, applied_upstream=None, bugs=None,
        authors=None, last_update=None):
    """Write a DEP-3 compliant patch.

    :param f: File-like object to write to
    :param repository: Repository to retrieve revisions from
    :param base_revid: Base revision id
    :param target_revid: Target revision id
    :param description: Optional description
    :param forwarded: Optional information on if/how the patch was forwarded
    :param applied_upstream: Optional information on how whether the patch
        was merged upstream
    :param bugs: Sequence of bug reports related to this patch
    :param authors: Sequence of authors of this patch
    :param last_update: Timestamp for last time this patch was updated
    """
    write_dep3_patch_header(f, bugs=bugs, authors=authors,
        last_update=last_update, description=description,
        revision_id=target_revid, origin=origin,
        applied_upstream=applied_upstream, forwarded=forwarded)
    old_tree = branch.repository.revision_tree(base_revid)
    new_tree = branch.repository.revision_tree(target_revid)
    diff.show_diff_trees(old_tree, new_tree, f, old_label='old/',
                         new_label='new/')



#    errors.py -- Error classes
#    Copyright (C) 2006 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from bzrlib.errors import BzrError


class DebianError(BzrError):
    _fmt = "A Debian packaging error occurred: %(cause)s"

    def __init__(self, cause):
        BzrError.__init__(self, cause=cause)


class NoSourceDirError(BzrError):
    _fmt = ("There is no existing source directory to use. Use "
            "--export-only or --dont-purge to get one that can be used")


class MissingUpstreamTarball(BzrError):
    _fmt = ("Unable to find the needed upstream tarball for package %(package)s, "
            "version %(version)s.")

    def __init__(self, package, version):
        BzrError.__init__(self, package=package, version=version)


class TarFailed(BzrError):
    _fmt = "There was an error executing tar to %(operation)s %(tarball)s."

    def __init__(self, operation, tarball):
        BzrError.__init__(self, operation=operation, tarball=tarball)


class BuildFailedError(BzrError):
    _fmt = "The build failed."


class UnparseableChangelog(BzrError):
    _fmt = "There was an error parsing the changelog: %(error)s"

    def __init__(self, error):
        BzrError.__init__(self, error=error)


class StopBuild(BzrError):
    _fmt = "Stopping the build: %(reason)s."

    def __init__(self, reason):
        BzrError.__init__(self, reason=reason)


class MissingChangelogError(BzrError):
    _fmt = 'Could not find changelog at %(location)s in tree.'

    def __init__(self, locations):
        BzrError.__init__(self, location=locations)


class AddChangelogError(BzrError):
    _fmt = 'Please add "%(changelog)s" to the branch using bzr add.'

    def __init__(self, changelog):
        BzrError.__init__(self, changelog=changelog)


class ImportError(BzrError):
    _fmt = "The files could not be imported: %(reason)s"

    def __init__(self, reason):
        BzrError.__init__(self, reason=reason)


class HookFailedError(BzrError):
    _fmt = 'The "%(hook_name)s" hook failed.'

    def __init__(self, hook_name):
        BzrError.__init__(self, hook_name=hook_name)


class OnlyImportSingleDsc(BzrError):
    _fmt = "You are only allowed to import one version in incremental mode."


class UnknownType(BzrError):
    _fmt = 'Cannot extract "%(path)s" from archive as it is an unknown type.'

    def __init__(self, path):
        BzrError.__init__(self, path=path)


class MissingChanges(BzrError):
    _fmt = "Could not find .changes file: %(changes)s."

    def __init__(self, changes):
        BzrError.__init__(self, changes=changes)


class UpstreamAlreadyImported(BzrError):
    _fmt = 'Upstream version "%(version)s" has already been imported.'

    def __init__(self, version):
        BzrError.__init__(self, version=str(version))


class UpstreamBranchAlreadyMerged(BzrError):
    _fmt = 'That revision of the upstream branch has already been merged.'


class AmbiguousPackageSpecification(BzrError):
    _fmt = ('You didn\'t specify a distribution with the package '
            'specification, and tags exists that state that the '
            'version that you specified has been uploaded to more '
            'than one distribution. Please specify which version '
            'you wish to refer to by by appending ":debian" or '
            '":ubuntu" to the revision specifier: %(specifier)s')

    def __init__(self, specifier):
        BzrError.__init__(self, specifier=specifier)


class UnknownVersion(BzrError):
    _fmt = ('No tag exists in this branch indicating that version '
            '"%(version)s" has been uploaded.')

    def __init__(self, version):
        BzrError.__init__(self, version=version)


class VersionNotSpecified(BzrError):
    _fmt = "You did not specify a package version."


class PackageVersionNotPresent(BzrError):
    _fmt = "%(package)s %(version)s was not found in %(upstream)s."

    def __init__(self, package, version, upstream):
        BzrError.__init__(self, package=package, version=version, 
                          upstream=upstream)


class UnsupportedRepackFormat(BzrError):
    _fmt = ('Either the file extension of "%(location)s" indicates that '
            'it is a format unsupported for repacking or it is a '
            'remote directory.')

    def __init__(self, location):
        BzrError.__init__(self, location=location)


class SharedUpstreamConflictsWithTargetPackaging(BzrError):

    _fmt = ('The upstream branches for the merge source and target have '
            'diverged. Unfortunately, the attempt to fix this problem '
            'resulted in conflicts. Please resolve these, commit and '
            're-run the "%(cmd)s" command to finish. '
            'Alternatively, until you commit you can use "bzr revert" to '
            'restore the state of the unmerged branch.')

    def __init__(self, cmd):
        self.cmd = cmd


class PerFileTimestampsNotSupported(BzrError):

    _fmt = ("Per file timestamps are not supported by the "
            "currently loaded version of bzrlib.")


class NoPreviousUpload(BzrError):

    _fmt = ("There was no previous upload to %(distribution)s.")

    def __init__(self, distribution):
        BzrError.__init__(self, distribution=distribution)


class UnableToFindPreviousUpload(BzrError):

    _fmt = ("Unable to determine the previous upload for --package-merge.")


class InconsistentSourceFormatError(BzrError):

    _fmt = ("Inconsistency between source format and version: version is "
            "%(version_bool)snative, format is %(format_bool)snative.")

    def __init__(self, version_native, format_native):
        if version_native:
            version_bool = ""
        else:
            version_bool = "not "
        if format_native:
            format_bool = ""
        else:
            format_bool = "not "
        BzrError.__init__(self, version_bool=version_bool, format_bool=format_bool)


class WatchFileMissing(BzrError):

    _fmt = "No watch file found."


class StrictBuildFailed(BzrError):

    _fmt = ("Build refused because there are unknown files in the tree. "
            "To list all known files, run 'bzr unknowns'.")


class DchError(BzrError):
    _fmt = 'There was an error using dch: %(error)s.'

    def __init__(self, error):
        BzrError.__init__(self, error=error)


class MultipleUpstreamTarballsNotSupported(BzrError):

    _fmt = ("Importing packages using source format 3.0 multiple tarballs "
            "is not yet supported.")


class QuiltUnapplyError(BzrError):

    _fmt = ("Unable to unapply quilt patches for %(kind)r tree: %(msg)s")

    def __init__(self, kind, msg):
        BzrError.__init__(self)
        self.kind = kind
        if msg is not None and msg.count("\n") == 1:
            msg = msg.strip()
        self.msg = msg



#    test_merge_upstream.py -- Blackbox tests for merge-upstream.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os.path

import bzrlib.export

from bzrlib.plugins.builddeb.tests import (
    BuilddebTestCase,
    SourcePackageBuilder,
    )
from bzrlib.plugins.builddeb.import_dsc import (
    DistributionBranch,
    DistributionBranchSet,
    )
from bzrlib.plugins.builddeb.util import (
    md5sum_filename,
    )


def iter_child_entries(tree, file_id):
    try:
        return tree.root_inventory.root.children.values()
    except AttributeError:
        return tree.inventory.root.children.values()


class Fixture(object):
    """A test fixture."""

    def __init__(self):
        pass

    def setUp(self, test_case):
        test_case.addCleanup(self.tearDown)

    def tearDown(self):
        pass


class Upstream(Fixture):
    """An upstream.

    :ivar tree: The tree of the upstream.
    """

    def setUp(self, test_case):
        Fixture.setUp(self, test_case)
        treename = test_case.getUniqueString()
        tree = test_case.make_branch_and_tree(treename)
        filename = test_case.getUniqueString()
        test_case.build_tree(["%s/%s" % (treename, filename)])
        tree.add([filename])
        tree.commit(test_case.getUniqueString())
        self.tree = tree


class ExportedTarball(Fixture):
    """An exported tarball 'release'."""

    def __init__(self, upstream, version):
        self.upstream = upstream
        self.version = version

    def setUp(self, test_case):
        filename = "project-%s.tar.gz" % self.version
        tree = self.upstream.tree.branch.repository.revision_tree(
            self.upstream.tree.branch.last_revision())
        bzrlib.export.export(tree, filename)
        self.tarball = filename


class DHMadePackage(Fixture):
    """A package made via dh-make."""

    def __init__(self, tar, upstream):
        self.tar = tar
        self.upstream = upstream

    def setUp(self, test_case):
        branchpath = test_case.getUniqueString()
        tree = self.upstream.tree.bzrdir.sprout(branchpath).open_workingtree()
        db = DistributionBranch(tree.branch, tree.branch, tree=tree,
                pristine_upstream_tree=tree)
        dbs = DistributionBranchSet()
        dbs.add_branch(db)
        db.import_upstream_tarballs(
            [(self.tar.tarball, None, md5sum_filename(self.tar.tarball))],
            "foo",
            str(self.tar.version),
            { None: [tree.branch.last_revision()]} )
        package_builder = SourcePackageBuilder("foo",
            str(self.tar.version)+"-1")
        package_builder.add_default_control()
        package_builder.write_debian_files(branchpath)
        tree.smart_add([tree.basedir])
        tree.commit('debianised.')
        self.tree = tree


class FileMovedReplacedUpstream(Fixture):
    """An upstream that has been changed by moving and replacing a file."""

    def __init__(self, upstream):
        self.upstream = upstream

    def setUp(self, test_case):
        branchpath = test_case.getUniqueString()
        tree = self.upstream.tree.bzrdir.sprout(branchpath).open_workingtree()
        self.tree = tree
        tree.lock_write()
        try:
            newpath = test_case.getUniqueString()
            for child in iter_child_entries(tree, tree.get_root_id()):
                if child.kind == 'file':
                    oldpath = child.name
            tree.rename_one(oldpath, newpath)
            test_case.build_tree(["%s/%s" % (os.path.basename(tree.basedir),
                oldpath)])
            tree.add([oldpath])
            tree.commit('yo, renaming and replacing')
        finally:
            tree.unlock()



class TestMergeUpstream(BuilddebTestCase):

    def test_merge_upstream_available(self):
        self.run_bzr('merge-upstream --help')

    def make_upstream(self):
        result = Upstream()
        result.setUp(self)
        return result

    def release_upstream(self, upstream):
        version = str(self.getUniqueInteger())
        upstream.tree.branch.tags.set_tag(version, upstream.tree.branch.last_revision())
        tar = ExportedTarball(upstream, version=version)
        tar.setUp(self)
        return tar

    def import_upstream(self, tar, upstream):
        packaging = DHMadePackage(tar, upstream)
        packaging.setUp(self)
        return packaging

    def file_moved_replaced_upstream(self, upstream):
        result = FileMovedReplacedUpstream(upstream)
        result.setUp(self)
        return result

    def test_smoke_renamed_file(self):
        # When a file is renamed by upstream, it should still import ok.
        upstream = self.make_upstream()
        rel1 = self.release_upstream(upstream)
        package = self.import_upstream(rel1, upstream)
        changed_upstream = self.file_moved_replaced_upstream(upstream)
        rel2 = self.release_upstream(changed_upstream)
        self.run_bzr(['merge-upstream', '--version', str(rel2.version),
            os.path.abspath(rel2.tarball)],
            working_dir=package.tree.basedir)

    def test_upstream_branch_revision_not_found(self):
        # When an upstream branch is specified but does not have the
        # upstream version, 'bzr merge-upstream' should complain.
        upstream = self.make_upstream()
        rel1 = self.release_upstream(upstream)
        package = self.import_upstream(rel1, upstream)
        changed_upstream = self.file_moved_replaced_upstream(upstream)
        rel2 = self.release_upstream(changed_upstream)
        self.run_bzr_error([
            'Using version string 8.',
            'bzr: ERROR: Version 8 can not be found in upstream branch <UpstreamBranchSource for \'.*\'>. Specify the revision manually using --revision or adjust \'export-upstream-revision\' in the configuration.'],
            ['merge-upstream', '--version', str(rel2.version),
            os.path.abspath(rel2.tarball), changed_upstream.tree.basedir],
            working_dir=package.tree.basedir)

    def test_upstream_branch_revision(self):
        # It is sufficient to just specify an upstream revision
        upstream = self.make_upstream()
        rel1 = self.release_upstream(upstream)
        package = self.import_upstream(rel1, upstream)
        changed_upstream = self.file_moved_replaced_upstream(upstream)
        rel2 = self.release_upstream(changed_upstream)
        package_path = package.tree.basedir
        os.mkdir(os.path.join(package_path, '.bzr-builddeb/'))
        f = open(os.path.join(package_path, '.bzr-builddeb/local.conf'), 'wb')
        try:
          f.write('[BUILDDEB]\nupstream-branch = %s\n' % changed_upstream.tree.basedir)
        finally:
          f.close()

        (out, err) = self.run_bzr(
            ['merge-upstream', '-rrevid:%s' % changed_upstream.tree.last_revision()],
            working_dir=package.tree.basedir)
        self.assertEquals(out, "")
        self.assertContainsRe(err, "Using version string 8.")

    def test_upstream_branch_revision_requires_upstream(self):
        # --revision requires a valid upstreambranch
        upstream = self.make_upstream()
        rel1 = self.release_upstream(upstream)
        package = self.import_upstream(rel1, upstream)
        changed_upstream = self.file_moved_replaced_upstream(upstream)
        rel2 = self.release_upstream(changed_upstream)
        (out, err) = self.run_bzr(['merge-upstream', '-r8'], working_dir=package.tree.basedir, retcode=3)
        self.assertEquals(out, "")
        self.assertEquals(err,
            "Using distribution unstable\n"
            "bzr: ERROR: --revision can only be used with a valid upstream branch\n")

    def test_hooks(self):
        upstream = self.make_upstream()
        rel1 = self.release_upstream(upstream)
        package = self.import_upstream(rel1, upstream)
        package_path = package.tree.basedir
        os.mkdir(os.path.join(package_path, '.bzr-builddeb/'))
        f = open(os.path.join(package_path, '.bzr-builddeb/local.conf'), 'wb')
        try:
          f.write('[HOOKS]\nmerge-upstream = touch muhook\n')
        finally:
          f.close()
        changed_upstream = self.file_moved_replaced_upstream(upstream)
        rel2 = self.release_upstream(changed_upstream)
        self.run_bzr(['merge-upstream', '--version', str(rel2.version),
            os.path.abspath(rel2.tarball)], working_dir=package.tree.basedir)
        self.assertPathExists(os.path.join(package.tree.basedir, 'muhook'))

    def test_new_package(self):
        upstream = self.make_upstream()
        tree = upstream.tree.bzrdir.sprout("package").open_workingtree()
        rel1 = self.release_upstream(upstream)
        self.run_bzr(['merge-upstream', '--version', str(rel1.version),
            "--package", "bar", os.path.abspath(rel1.tarball)],
            working_dir=tree.basedir)

    def test_new_package_from_empty_branch(self):
        upstream = self.make_upstream()
        tree = self.make_branch_and_tree("package")
        rel1 = self.release_upstream(upstream)
        self.run_bzr(['merge-upstream', '--version', str(rel1.version),
            "--package", "bar", os.path.abspath(rel1.tarball)],
            working_dir=tree.basedir)

# vim: ts=4 sts=4 sw=4



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright ? 2008 Canonical Ltd.
# Author: Scott James Remnant <scott@ubuntu.com>.
# Hacked up by: Bryce Harrington <bryce@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import logging
import os.path
import re
import shutil
import subprocess
import tempfile

from bzrlib import (
    merge,
    osutils,
    )


# A logger in the 'bzr' hierarchy.  By default messages will be propagated to
# the standard bzr logger, but tests can easily intercept just this logger if
# they wish.
_logger = logging.getLogger('bzr.plugins.builddeb.merge_changelog')


class ChangeLogFileMerge(merge.ConfigurableFileMerger):

    name_prefix = 'deb_changelog'
    default_files = ['debian/changelog']

    def merge_text(self, params):
        return merge_changelog(params.this_lines, params.other_lines,
            params.base_lines)


def merge_changelog(this_lines, other_lines, base_lines=[]):
    """Merge a changelog file."""
    # Write the BASE, THIS and OTHER versions to files in a temporary
    # directory, and use dpkg-mergechangelogs to merge them.
    tmpdir = tempfile.mkdtemp('deb_changelog_merge')
    try:
        def writelines(filename, lines):
            with open(filename, 'w') as f:
                for line in lines:
                    f.write(line)
        base_filename = os.path.join(tmpdir, 'changelog.base')
        this_filename = os.path.join(tmpdir, 'changelog.this')
        other_filename = os.path.join(tmpdir, 'changelog.other')
        writelines(base_filename, base_lines)
        writelines(this_filename, this_lines)
        writelines(other_filename, other_lines)
        try:
            proc = subprocess.Popen(['dpkg-mergechangelogs', base_filename,
                this_filename, other_filename], stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        except OSError, e:
            if e.errno == errno.ENOENT:
                # No dpkg-mergechangelogs command available
                return 'not_applicable', ''
            raise
        stdout, stderr = proc.communicate()
        retcode = proc.returncode
        if stderr:
            # Relay the warning from dpkg-mergechangelogs to the user.  We
            # don't decorate the messages at all, as dpkg-mergechangelogs
            # warnings are already prefixed with "dpkg-mergechangelogs:
            # warning:" which makes the origin of the messages quite clear.
            encoding = osutils.get_user_encoding()
            # Errors are output using the locale, and log needs unicode.
            _logger.warning('%s', stderr.decode(encoding, "replace"))
        if retcode == 1:
            # dpkg-mergechangelogs reports a conflict.  Unfortunately it uses
            # slightly non-standard conflict markers (<http://pad.lv/815700>:
            # "<<<<<<" rather than "<<<<<<<", i.e. 6 chars instead of 7), so we
            # correct that here to make the results of this plugin as
            # consistent with regular bzr usage as possible.  Note that
            # conflict markers are never valid lines in a changelog file, so
            # it's reasonable for us to assume that any line that looks like a
            # conflict marker is a conflict marker (rather than valid content).
            # At worst a conflicted merge of an invalid changelog file that
            # already contained a non-standard conflict marker will have that
            # conflict marker made standard, which is more like a feature than
            # a bug!
            def replace_func(match_obj):
                match_text = match_obj.group(0)
                return match_text[0] * 7
            stdout = re.sub('(?m)^[<=>]{6}$', replace_func, stdout)
            return 'conflicted', stdout
        elif retcode != 0:
            # dpkg-mergechangelogs exited with an error. There is probably no
            # output at all, but regardless the merge should fall back to
            # another method.
            _logger.warning("dpkg-mergechangelogs failed with status %d", retcode)
            return 'not_applicable', stdout
        else:
            return 'success', stdout
    finally:
        shutil.rmtree(tmpdir)



#    import_dsc.py -- Import a series of .dsc files.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#              (C) 2008 Canonical Ltd.
#
#    Code is also taken from bzrtools, which is
#             (C) 2005, 2006, 2007 Aaron Bentley <aaron.bentley@utoronto.ca>
#             (C) 2005, 2006 Canonical Limited.
#             (C) 2006 Michael Ellerman.
#    and distributed under the GPL, version 2 or later.
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


import calendar
import os
import shutil
import stat
import subprocess
import tempfile

try:
    from debian import deb822
    from debian.changelog import Version, Changelog, VersionError
except ImportError:
    # Prior to 0.1.15 the debian module was called debian_bundle
    from debian_bundle import deb822
    from debian_bundle.changelog import Version, Changelog, VersionError

from bzrlib import (
                    bzrdir,
                    osutils,
                    )
from bzrlib.config import ConfigObj
from bzrlib.errors import (
        AlreadyBranchError,
        BzrCommandError,
        NotBranchError,
        NoWorkingTree,
        UnrelatedBranches,
        )
from bzrlib.revision import NULL_REVISION
from bzrlib.trace import warning, mutter
from bzrlib.transport import (
    get_transport,
    )

from bzrlib.plugins.builddeb.bzrtools_import import import_dir
from bzrlib.plugins.builddeb.errors import (
    MultipleUpstreamTarballsNotSupported,
    PackageVersionNotPresent,
    UpstreamAlreadyImported,
    UpstreamBranchAlreadyMerged,
    )
from bzrlib.plugins.builddeb.util import (
    FORMAT_1_0,
    FORMAT_3_0_QUILT,
    FORMAT_3_0_NATIVE,
    component_from_orig_tarball,
    export,
    extract_orig_tarballs,
    get_commit_info_from_changelog,
    md5sum_filename,
    open_file_via_transport,
    open_transport,
    safe_decode,
    subprocess_setup,
    )
from bzrlib.plugins.builddeb.upstream.pristinetar import (
    PristineTarSource,
    )


class DscCache(object):

    def __init__(self, transport=None):
        self.cache = {}
        self.transport_cache = {}
        self.transport = transport

    def get_dsc(self, name):

        if name in self.cache:
            dsc1 = self.cache[name]
        else:
            # Obtain the dsc file, following any redirects as needed.
            filename, transport = open_transport(name)
            f1 = open_file_via_transport(filename, transport)
            try:
                dsc1 = deb822.Dsc(f1)
            finally:
                f1.close()
            self.cache[name] = dsc1
            self.transport_cache[name] = transport

        return dsc1

    def get_transport(self, name):
        return self.transport_cache[name]


class DscComp(object):

    def __init__(self, cache):
        self.cache = cache

    def cmp(self, dscname1, dscname2):
        dsc1 = self.cache.get_dsc(dscname1)
        dsc2 = self.cache.get_dsc(dscname2)
        v1 = Version(dsc1['Version'])
        v2 = Version(dsc2['Version'])
        if v1 == v2:
            return 0
        if v1 > v2:
            return 1
        return -1


class DistributionBranchSet(object):
    """A collection of DistributionBranches with an ordering.

    A DistributionBranchSet collects a group of DistributionBranches
    and an order, and then can provide the branches with information
    about their place in the relationship with other branches.
    """

    def __init__(self):
        """Create a DistributionBranchSet."""
        self._branch_list = []

    def add_branch(self, branch):
        """Adds a DistributionBranch to the end of the list.

        Appends the passed distribution branch to the end of the list
        that this DistributionBranchSet represents. It also provides
        the distribution branch with a way to get the branches that
        are before and after it in the list.

        It will call branch.set_get_lesser_branches_callback() and
        branch.set_get_greater_branches_callback(), passing it methods
        that the DistributionBranch can call to get the list of branches
        before it in the list and after it in the list respectively.
        The passed methods take no arguments and return a list (possibly
        empty) of the desired branches.

        :param branch: the DistributionBranch to add.
        """
        self._branch_list.append(branch)
        lesser_callback = self._make_lesser_callback(branch)
        branch.set_get_lesser_branches_callback(lesser_callback)
        greater_callback = self._make_greater_callback(branch)
        branch.set_get_greater_branches_callback(greater_callback)

    def _make_lesser_callback(self, branch):
        return lambda: self.get_lesser_branches(branch)

    def _make_greater_callback(self, branch):
        return lambda: self.get_greater_branches(branch)

    def get_lesser_branches(self, branch):
        """Return the list of branches less than the argument.

        :param branch: The branch that all branches returned must be less
            than.
        :return: a (possibly empty) list of all the branches that are
            less than the argument. The list is sorted starting with the
            least element.
        """
        index = self._branch_list.index(branch)
        return self._branch_list[:index]

    def get_greater_branches(self, branch):
        """Return the list of branches greater than the argument.

        :param branch: The branch that all branches returned must be greater
            than.
        :return: a (possibly empty) list of all the branches that are
            greater than the argument. The list is sorted starting with the
            least element.
        """
        index = self._branch_list.index(branch)
        return self._branch_list[index+1:]


def checkout_upstream_version(tree, package, version, revisions):
    """Checkout an upstream version from the pristine tar source.

    """
    tree.update(revision=revisions[None])
    parent_ids = []
    for component in sorted(revisions.keys()):
        revid = revisions[component]
        if component is not None:
            component_tree = tree.branch.repository.revision_tree(revid)
            export(component_tree, os.path.join(tree.basedir, component), format='dir')
        parent_ids.append(revid)
    tree.set_parent_ids(parent_ids)


class DistributionBranch(object):
    """A DistributionBranch is a representation of one line of development.

    It is a branch that is linked to a line of development, such as Debian
    unstable. It also has associated branches, some of which are "lesser"
    and some are "greater". A lesser branch is one that this branch
    derives from. A greater branch is one that derives from this. For
    instance Debian experimental would have unstable as a lesser branch,
    and vice-versa. It is assumed that a group of DistributionBranches will
    have a total ordering with respect to these relationships.
    """

    def __init__(self, branch, pristine_upstream_branch, tree=None,
                 pristine_upstream_tree=None):
        """Create a distribution branch.

        You can only import packages on to the DistributionBranch
        if both tree and pristine_upstream_tree are provided.

        :param branch: the Branch for the packaging part.
        :param pristine_upstream_branch: the Branch for the pristine tar part,
            if any.
        :param tree: an optional tree for the branch.
        :param pristine_upstream_tree: an optional tree for the
            pristine_upstream_branch.
        """
        self.branch = branch
        self.tree = tree
        self.pristine_upstream_branch = pristine_upstream_branch
        self.pristine_upstream_tree = pristine_upstream_tree
        if pristine_upstream_branch is not None:
            self.pristine_upstream_source = PristineTarSource(
                branch=pristine_upstream_branch, tree=pristine_upstream_tree)
        else:
            self.pristine_upstream_source = None
        self.get_lesser_branches = None
        self.get_greater_branches = None

    def set_get_lesser_branches_callback(self, callback):
        """Set the callback to get the branches "lesser" than this.

        The function passed to this method will be used to get the
        list of branches that are "lesser" than this one. It is
        expected to require no arguments, and to return the desired
        (possibly empty) list of branches. The returned list should
        be sorted starting with the least element.

        :param callback: a function that is called to get the desired list
            of branches.
        """
        self.get_lesser_branches = callback

    def set_get_greater_branches_callback(self, callback):
        """Set the callback to get the branches "greater" than this.

        The function passed to this method will be used to get the
        list of branches that are "greater" than this one. It is
        expected to require no arguments, and to return the desired
        (possibly empty) list of branches. The returned list should
        be sorted starting with the least element.

        :param callback: a function that is called to get the desired list
            of branches.
        """
        self.get_greater_branches = callback

    def get_other_branches(self):
        """Return all the other branches in this set.

        The returned list will be ordered, and will not contain this
        branch.

        :return: a list of all the other branches in this set (if any).
        """
        return self.get_lesser_branches() + self.get_greater_branches()

    def tag_name(self, version):
        """Gets the name of the tag that is used for the version.

        :param version: the Version object that the tag should refer to.
        :return: a String with the name of the tag.
        """
        return str(version)

    def _has_version(self, branch, tag_name, md5=None):
        if not branch.tags.has_tag(tag_name):
            return False
        revid = branch.tags.lookup_tag(tag_name)
        branch.lock_read()
        try:
            graph = branch.repository.get_graph()
            if not graph.is_ancestor(revid, branch.last_revision()):
                return False
        finally:
            branch.unlock()
        if md5 is None:
            return True
        rev = branch.repository.get_revision(revid)
        try:
            return rev.properties['deb-md5'] == md5
        except KeyError:
            warning("tag %s present in branch, but there is no "
                "associated 'deb-md5' property" % tag_name)
            return False

    def has_version(self, version, md5=None):
        """Whether this branch contains the package version specified.

        The version must be judged present by having the appropriate tag
        in the branch. If the md5 argument is not None then the string
        passed must the the md5sum that is associated with the revision
        pointed to by the tag.

        :param version: a Version object to look for in this branch.
        :param md5: a string with the md5sum that if not None must be
            associated with the revision.
        :return: True if this branch contains the specified version of the
            package. False otherwise.
        """
        tag_name = self.tag_name(version)
        if self._has_version(self.branch, tag_name, md5=md5):
            return True
        debian_tag_name = "debian-" + tag_name
        if self._has_version(self.branch, debian_tag_name, md5=md5):
            return True
        ubuntu_tag_name = "ubuntu-" + tag_name
        if self._has_version(self.branch, ubuntu_tag_name, md5=md5):
            return True
        return False

    def contained_versions(self, versions):
        """Splits a list of versions depending on presence in the branch.

        Partitions the input list of versions depending on whether they
        are present in the branch or not.

        The two output lists will be sorted in the same order as the input
        list.

        :param versions: a list of Version objects to look for in the
            branch. May be an empty list.
        :return: A tuple of two lists. The first list is the list of those
            items from the input list that are present in the branch. The
            second list is the list of those items from the input list that
            are not present in the branch. The two lists will be disjoint
            and cover the input list. Either list may be empty, or both if
            the input list is empty.
        """
        #FIXME: should probably do an ancestory check to find all
        # merged revisions. This will avoid adding an extra parent
        # when say
        # experimental 1-1~rc1
        # unstable 1-1 1-1~rc1
        # Ubuntu 1-1ubuntu1 1-1 1-1~rc1
        # where only the first in each list is actually uploaded.
        contained = []
        not_contained = []
        for version in versions:
            if self.has_version(version):
                contained.append(version)
            else:
                not_contained.append(version)
        return contained, not_contained

    def missing_versions(self, versions):
        """Returns the versions from the list that the branch does not have.

        Looks at all the versions specified and returns a list of the ones
        that are earlier in the list that the last version that is
        contained in this branch.

        :param versions: a list of Version objects to look for in the branch.
            May be an empty list.
        :return: The subset of versions from the list that are not present
            in this branch. May be an empty list.
        """
        last_contained = self.last_contained_version(versions)
        if last_contained is None:
            return versions
        index = versions.index(last_contained)
        return versions[:index]

    def last_contained_version(self, versions):
        """Returns the highest version from the list present in this branch.

        It assumes that the input list of versions is sorted with the
        highest version first.

        :param versions: a list of Version objects to look for in the branch.
            Must be sorted with the highest version first. May be an empty
            list.
        :return: the highest version that is contained in this branch, or
            None if none of the versions are contained within the branch.
        """
        for version in versions:
            if self.has_version(version):
                return version
        return None

    def revid_of_version(self, version):
        """Returns the revision id corresponding to that version.

        :param version: the Version object that you wish to retrieve the
            revision id of. The Version must be present in the branch.
        :return: the revision id corresponding to that version
        """
        tag_name = self.tag_name(version)
        if self._has_version(self.branch, tag_name):
            return self.branch.tags.lookup_tag(tag_name)
        debian_tag_name = "debian-" + tag_name
        if self._has_version(self.branch, debian_tag_name):
            return self.branch.tags.lookup_tag(debian_tag_name)
        ubuntu_tag_name = "ubuntu-" + tag_name
        if self._has_version(self.branch, ubuntu_tag_name):
            return self.branch.tags.lookup_tag(ubuntu_tag_name)
        return self.branch.tags.lookup_tag(tag_name)

    def tag_version(self, version, revid=None):
        """Tags the branch's last revision with the given version.

        Sets a tag on the last revision of the branch with a tag that refers
        to the version provided.

        :param version: the Version object to derive the tag name from.
        :param revid: the revid to associate the tag with, or None for the
            tip of self.branch.
        :return: Name of the tag set
        """
        tag_name = self.tag_name(version)
        if revid is None:
            revid = self.branch.last_revision()
        self.branch.tags.set_tag(tag_name, revid)
        return tag_name


    def _default_config_for_tree(self, tree):
        # FIXME: shouldn't go to configobj directly
        for path in ('debian/bzr-builddeb.conf', '.bzr-builddeb/default.conf',):
            c_fileid = tree.path2id(path)
            if c_fileid is not None:
                break
        else:
            return None, None, None
        tree.lock_read()
        try:
            config = ConfigObj(tree.get_file(c_fileid, path))
            try:
                config['BUILDDEB']
            except KeyError:
                config['BUILDDEB'] = {}
        finally:
            tree.unlock()
        return c_fileid, path, config

    def _is_tree_native(self, config):
        if config is not None:
            try:
                current_value = config['BUILDDEB']['native']
            except KeyError:
                current_value = False
            return current_value == "True"
        return False

    def is_version_native(self, version):
        """Determines whether the given version is native.

        :param version: the Version object to test. Must be present in
            the branch.
        :return: True if the version is was recorded as native when
            imported, False otherwise.
        """
        revid = self.revid_of_version(version)
        rev_tree = self.branch.repository.revision_tree(revid)
        (config_fileid, config_path,
         current_config) = self._default_config_for_tree(rev_tree)
        if self._is_tree_native(current_config):
            return True
        rev = self.branch.repository.get_revision(revid)
        try:
            prop = rev.properties["deb-native"]
            return prop == "True"
        except KeyError:
            return False

    def can_pull_from_branch(self, branch, version, md5):
        if not branch.has_version(version, md5=md5):
            return False

        # Check that they haven't diverged
        branch.branch.lock_read()
        try:
            graph = branch.branch.repository.get_graph(
                    self.branch.repository)
            return graph.is_ancestor(self.branch.last_revision(),
                    branch.revid_of_version(version))
        finally:
            branch.branch.unlock()


    def branch_to_pull_version_from(self, version, md5):
        """Checks whether this upload is a pull from a lesser branch.

        Looks in all the lesser branches for the given version/md5 pair
        in a branch that has not diverged from this.

        If it is present in another branch that has not diverged this
        method will return the greatest branch that it is present in,
        otherwise it will return None. If it returns a branch then it
        indicates that a pull should be done from that branch, rather
        than importing the version as a new revision in this branch.

        :param version: the Version object to look for in the lesser
            branches.
        :param md5: a String containing the md5 associateed with the
            version.
        :return: a DistributionBranch object to pull from if that is
            what should be done, otherwise None.
        """
        assert md5 is not None, \
            ("It's not a good idea to use branch_to_pull_version_from with "
             "md5 == None, as you may pull the wrong revision.")
        self.branch.lock_read()
        try:
            for branch in reversed(self.get_lesser_branches()):
                if self.can_pull_from_branch(branch, version, md5):
                    return branch
            for branch in self.get_greater_branches():
                if self.can_pull_from_branch(branch, version, md5):
                    return branch
            return None
        finally:
            self.branch.unlock()

    def can_pull_upstream_from_branch(self, branch, package, version,
            upstream_tarballs=None):
        """Check if a version can be pulled from another branch into this one.

        :param branch: Branch with upstream version
        :param package: Package name
        :param version: Package version
        :param upstream_tarballs: Required upstream tarballs (optional)
        """
        if not branch.pristine_upstream_source.has_version(package, version,
                tarballs=upstream_tarballs):
            return False

        up_branch = self.pristine_upstream_branch
        up_branch.lock_read()
        try:
            # Check that they haven't diverged
            other_up_branch = branch.pristine_upstream_branch
            other_up_branch.lock_read()
            try:
                graph = other_up_branch.repository.get_graph(
                        up_branch.repository)
                pristine_upstream_revids = branch.pristine_upstream_source.version_as_revisions(
                        package, version, tarballs=upstream_tarballs)
                for (component, pristine_upstream_revid) in pristine_upstream_revids.iteritems():
                    if not graph.is_ancestor(up_branch.last_revision(),
                            pristine_upstream_revid):
                        return False
                return True
            finally:
                other_up_branch.unlock()
        finally:
            up_branch.unlock()

    def branch_to_pull_upstream_from(self, package, version, upstream_tarballs):
        """Checks whether this upstream is a pull from a lesser branch.

        Looks in all the other upstream branches for the given
        version/md5 pair in a branch that has not diverged from this.
        If it is present in a lower branch this method will return the
        greatest branch that it is present in that has not diverged,
        otherwise it will return None. If it returns a branch then it
        indicates that a pull should be done from that branch, rather
        than importing the upstream as a new revision in this branch.

        :param version: the upstream version to use when searching in the 
            lesser branches.
        :return: a DistributionBranch object to pull the upstream from
            if that is what should be done, otherwise None.
        """
        assert isinstance(version, str)
        for branch in reversed(self.get_lesser_branches()):
            if self.can_pull_upstream_from_branch(branch, package, version,
                    upstream_tarballs):
                return branch
        for branch in self.get_greater_branches():
            if self.can_pull_upstream_from_branch(branch, package, version,
                    upstream_tarballs):
                return branch
        return None

    def get_parents(self, versions):
        """Return the list of parents for a specific version.

        This method returns the list of revision ids that should be parents
        for importing a specific package version. The specific package version
        is the first element of the list of versions passed.

        The parents are determined by looking at the other versions in the
        passed list and examining which of the branches (if any) they are
        already present in.

        You should probably use get_parents_with_upstream rather than
        this method.

        :param versions: a list of Version objects, the first item of
            which is the version of the package that is currently being
            imported.
        :return: a list of tuples of (DistributionBranch, version,
            revision id). The revision ids should all be parents of the
            revision that imports the specified version of the package.
            The versions are the versions that correspond to that revision
            id. The DistributionBranch is the branch that contains that
            version.
        """
        assert len(versions) > 0, "Need a version to import"
        mutter("Getting parents of %s" % str(versions))
        missing_versions = self.missing_versions(versions)
        mutter("Versions we don't have are %s" % str(missing_versions))
        last_contained_version = self.last_contained_version(versions)
        parents = []
        if last_contained_version is not None:
            assert last_contained_version != versions[0], \
                "Reupload of a version?"
            mutter("The last versions we do have is %s" \
                    % str(last_contained_version))
            parents = [(self, last_contained_version,
                    self.revid_of_version(last_contained_version))]
        else:
            mutter("We don't have any of those versions")
        for branch in list(reversed(self.get_lesser_branches())) + self.get_greater_branches():
            merged, missing_versions = \
                branch.contained_versions(missing_versions)
            if merged:
                revid = branch.revid_of_version(merged[0])
                parents.append((branch, merged[0], revid))
                mutter("Adding merge from related branch of %s for version %s"
                        % (revid, str(merged[0])))
                #FIXME: should this really be here?
                self._fetch_from_branch(branch, revid)
        return parents

    def pull_upstream_from_branch(self, pull_branch, package, version):
        """Pulls an upstream version from a branch.

        Given a DistributionBranch and a version number this method
        will pull the upstream part of the given version from the
        branch in to this. The upstream version must be present
        in the DistributionBranch, and it is assumed that the md5
        matches.

        It sets the necessary tags so that the pulled version is
        recognised as being part of this branch.

        :param pull_branch: the DistributionBranch to pull from.
        :param version: the upstream version string
        """
        assert isinstance(version, str)
        pull_revisions = pull_branch.pristine_upstream_source.version_as_revisions(
            package, version)
        for (component, pull_revision) in pull_revisions.iteritems():
            mutter("Fetching upstream part %s of %s from revision %s" % \
                    (component, version, pull_revision))
            assert self.pristine_upstream_tree is not None, \
                "Can't pull upstream with no tree"
            self.pristine_upstream_branch.pull(pull_branch.pristine_upstream_branch,
                    stop_revision=pull_revision)
            self.pristine_upstream_source.tag_version(version, pull_revision)
            self.branch.fetch(self.pristine_upstream_branch, last_revision=pull_revision)
            self.pristine_upstream_branch.tags.merge_to(self.branch.tags)
        checkout_upstream_version(self.pristine_upstream_tree,
            package, version, pull_revisions)

    def pull_version_from_branch(self, pull_branch, package, version, native=False):
        """Pull a version from a particular branch.

        Given a DistributionBranch and a version number this method
        will pull the given version from the branch in to this. The
        version must be present in the DistributionBranch, and it
        is assumed that the md5 matches.

        It will also pull in any upstream part that is needed to
        the upstream branch. It is assumed that the md5 matches
        here as well. If the upstream version must be present in
        at least one of the upstream branches.

        It sets the necessary tags on the revisions so they are
        recongnised in this branch as well.

        :param pull_branch: the DistributionBranch to pull from.
        :param version: the Version to pull.
        :param native: whether it is a native version that is being
            imported.
        """
        pull_revision = pull_branch.revid_of_version(version)
        mutter("already has version %s so pulling from revision %s"
                % (str(version), pull_revision))
        assert self.tree is not None, "Can't pull branch with no tree"
        self.tree.pull(pull_branch.branch, stop_revision=pull_revision)
        self.tag_version(version, revid=pull_revision)
        if not native and not self.pristine_upstream_source.has_version(package, version.upstream_version):
            if pull_branch.pristine_upstream_source.has_version(package, version.upstream_version):
                self.pull_upstream_from_branch(pull_branch, 
                    package, version.upstream_version)
            else:
                assert False, ("Can't find the needed upstream part "
                        "for version %s" % version)
        if (native and self.pristine_upstream_branch.last_revision() == NULL_REVISION
            and pull_branch.pristine_upstream_branch.last_revision() != NULL_REVISION):
            # in case the package wasn't native before then we pull
            # the upstream. These checks may be a bit restrictive.
            self.pristine_upstream_tree.pull(pull_branch.pristine_upstream_branch)
            pull_branch.pristine_upstream_branch.tags.merge_to(self.pristine_upstream_branch.tags)
        elif native:
            mutter("Not checking for upstream as it is a native package")
        else:
            mutter("Not importing the upstream part as it is already "
                    "present in the upstream branch")

    def get_parents_with_upstream(self, package, version, versions,
            tarballs, force_upstream_parent=False):
        """Get the list of parents including any upstream parents.

        Further to get_parents this method includes any upstream parents
        that are needed. An upstream parent is needed if none of
        the other parents include the upstream version. The needed
        upstream must already present in the upstream branch before
        calling this method.

        If force_upstream_parent is True then the upstream parent will
        be included, even if another parent is already using that
        upstream. This is for use in cases where the .orig.tar.gz
        is different in two distributions.

        :param version: the Version that we are currently importing.
        :param versions: the list of Versions that are ancestors of
            version, including version itself. Sorted with the latest
            versions first, so version must be the first entry.
        :param force_upstream_parent: if True then an upstream parent
            will be added as the first parent, regardless of what the
            other parents are.
        :return: a list of revision ids that should be the parents when
            importing the specified revision.
        """
        assert version == versions[0], \
            "version is not the first entry of versions"
        parents = self.get_parents(versions)
        need_upstream_parent = True
        if not force_upstream_parent:
            for parent_pair in parents:
                if (parent_pair[1].upstream_version == \
                        version.upstream_version):
                    need_upstream_parent = False
                    break
        real_parents = [p[2] for p in parents]
        if need_upstream_parent:
            upstream_revids = self.pristine_upstream_source.version_as_revisions(
                package, version.upstream_version, tarballs)
            def key(a):
                if a is None:
                    return None
                return a
            for component in sorted(upstream_revids.keys(), key=key):
                if len(real_parents) > 0:
                    real_parents.insert(1, upstream_revids[component])
                else:
                    real_parents = [upstream_revids[component]]
        return real_parents

    def _fetch_upstream_to_branch(self, imported_revids):
        """Fetch the revision from the upstream branch in to the packaging one.
        """
        # Make sure we see any revisions added by the upstream branch
        # since self.tree was locked.
        self.branch.repository.refresh_data()
        for (component, tag, revid) in imported_revids:
            self.branch.fetch(self.pristine_upstream_branch, last_revision=revid)
        self.pristine_upstream_branch.tags.merge_to(self.branch.tags)

    def import_upstream(self, upstream_part, package, version, upstream_parents,
            upstream_tarballs, upstream_branch=None,
            upstream_revisions=None, timestamp=None, author=None,
            file_ids_from=None):
        """Import an upstream part on to the upstream branch.

        This imports the upstream part of the code and places it on to
        the upstream branch, setting the necessary tags.

        :param upstream_part: the path of a directory containing the
            unpacked upstream part of the source package.
        :param version: upstream version that is being imported
        :param upstream_parents: the parents to give the upstream revision
        :param timestamp: a tuple of (timestamp, timezone) to use for
            the commit, or None to use the current time.
        :return: list with (component, tag, revid) tuples
        """
        # Should we just dump the upstream part on whatever is currently
        # there, or try and pull all of the other upstream versions
        # from lesser branches first? For now we'll just dump it on.
        # TODO: this method needs a lot of work for when we will make
        # the branches writeable by others.
        assert isinstance(version, str)
        mutter("Importing upstream version %s from %s with parents %r" \
                % (version, upstream_part, upstream_parents))
        assert self.pristine_upstream_tree is not None, \
            "Can't import upstream with no tree"
        other_branches = self.get_other_branches()
        ret = []
        for (tarball, component, md5) in upstream_tarballs:
            parents = upstream_parents.get(component, [])
            if upstream_revisions is not None:
                revid = upstream_revisions[component]
            else:
                revid = None
            upstream_trees = [o.pristine_upstream_branch.basis_tree()
                for o in other_branches]
            target_tree = None
            if upstream_branch is not None:
                if revid is None:
                    # FIXME: This is wrong for component tarballs
                    revid = upstream_branch.last_revision()
                self.pristine_upstream_branch.fetch(upstream_branch,
                        last_revision=revid)
                upstream_branch.tags.merge_to(self.pristine_upstream_branch.tags)
                parents.append(revid)
                target_tree = self.pristine_upstream_branch.repository.revision_tree(
                            revid)
            if file_ids_from is not None:
                upstream_trees = file_ids_from + upstream_trees
            if self.tree:
                self_tree = self.tree
                self_tree.lock_write() # might also be upstream tree for dh_make
            else:
                self_tree = self.branch.basis_tree()
                self_tree.lock_read()
            if len(parents) > 0:
                parent_revid = parents[0]
            else:
                parent_revid = NULL_REVISION
            self.pristine_upstream_tree.pull(self.pristine_upstream_tree.branch,
                overwrite=True, stop_revision=parent_revid)
            if component is None:
                path = upstream_part
            else:
                path = os.path.join(upstream_part, component)
            try:
                import_dir(self.pristine_upstream_tree, path,
                        file_ids_from=[self_tree] + upstream_trees,
                        target_tree=target_tree)
            finally:
                self_tree.unlock()
            if component is None:
                exclude = [tb[1] for tb in upstream_tarballs if tb[1] is not None]
            else:
                exclude = []
            (tag, revid) = self.pristine_upstream_source.import_component_tarball(
                package, version, self.pristine_upstream_tree, parents,
                component, md5, tarball, author=author, timestamp=timestamp,
                exclude=exclude)
            self.pristine_upstream_branch.generate_revision_history(revid)
            ret.append((component, tag, revid))
            self.branch.fetch(self.pristine_upstream_branch)
            self.branch.tags.set_tag(tag, revid)
        return ret

    def import_upstream_tarballs(self, tarballs, package, version, parents,
        upstream_branch=None, upstream_revisions=None):
        """Import an upstream part to the upstream branch.

        :param tarballs: List of tarballs / components to extract
        :param version: The upstream version to import.
        :param parents: The tarball-branch parents to use for the import.
            If an upstream branch is supplied, its automatically added to
            parents.
        :param upstream_branch: An upstream branch to associate with the
            tarball.
        :param upstream_revisions: Upstream revision ids dictionary
        :param md5sum: hex digest of the md5sum of the tarball, if known.
        :return: list with (component, tag, revid) tuples
        """
        tarball_dir = self._extract_tarballs_to_tempdir(tarballs)
        try:
            return self.import_upstream(tarball_dir, package, version, parents,
                tarballs,
                upstream_branch=upstream_branch,
                upstream_revisions=upstream_revisions)
        finally:
            shutil.rmtree(tarball_dir)

    def _mark_native_config(self, native):
        poss_native_tree = self.branch.basis_tree()
        (config_fileid, config_relpath,
         current_config) = self._default_config_for_tree(poss_native_tree)
        current_native = self._is_tree_native(current_config)
        if current_config is not None:
            # Add that back to the current tree
            current_config.filename = self.tree.abspath(config_relpath)
            dir_path = osutils.dirname(current_config.filename)
            if not os.path.exists(dir_path):
                os.mkdir(dir_path)
            current_config.write()
            dirname = osutils.dirname(config_relpath)
            dir_id = poss_native_tree.path2id(dirname)
            self.tree.add([dirname, config_relpath],
                          ids=[dir_id, config_fileid])
        if native != current_native:
            if current_config is None:
                needs_add = True
                if native:
                    current_config = ConfigObj()
                    current_config['BUILDDEB'] = {}
            if current_config is not None:
                if native:
                    current_config['BUILDDEB']['native'] = True
                else:
                    del current_config['BUILDDEB']['native']
                    if len(current_config['BUILDDEB']) == 0:
                        del current_config['BUILDDEB']
                if len(current_config) == 0:
                    self.tree.remove(['.bzr-builddeb',
                            '.bzr-builddeb/default.conf',
                            'debian/bzr-builddeb.conf'],
                            keep_files=False)
                else:
                    current_config.filename = os.path.join(
                        self.tree.basedir, 'debian', 'bzr-builddeb.conf')
                    current_config.write()
                    if needs_add:
                        self.tree.add(['debian', 'debian/bzr-builddeb.conf'])

    def import_debian(self, debian_part, version, parents, md5,
            native=False, timestamp=None, file_ids_from=None):
        """Import the debian part of a source package.

        :param debian_part: the path of a directory containing the unpacked
            source package.
        :param version: the Version of the source package.
        :param parents: a list of revision ids that should be the
            parents of the imported revision.
        :param md5: the md5 sum reported by the .dsc for
            the .diff.gz part of this source package.
        :param native: whether the package is native.
        :param timestamp: a tuple of (timestamp, timezone) to use for
            the commit, or None to use the current values.
        """
        mutter("Importing debian part for version %s from %s, with parents "
                "%s" % (str(version), debian_part, str(parents)))
        assert self.tree is not None, "Can't import with no tree"
        # First we move the branch to the first parent
        if parents:
            if self.branch.last_revision() == NULL_REVISION:
                parent_revid = parents[0]
                self.tree.pull(self.tree.branch, overwrite=True,
                        stop_revision=parent_revid)
            elif parents[0] != self.branch.last_revision():
                mutter("Adding current tip as parent: %s"
                        % self.branch.last_revision())
                parents.insert(0, self.branch.last_revision())
        elif self.branch.last_revision() != NULL_REVISION:
            # We were told to import with no parents. That's not
            # right, so import with the current parent. Should
            # perhaps be fixed in the methods to determine the parents.
            mutter("Told to import with no parents. Adding current tip "
                   "as the single parent")
            parents = [self.branch.last_revision()]
        other_branches = self.get_other_branches()
        debian_trees = [o.branch.basis_tree() for o in other_branches]
        parent_trees = []
        if file_ids_from is not None:
            parent_trees = file_ids_from[:]
        for parent in parents:
            parent_trees.append(self.branch.repository.revision_tree(
                        parent))
        import_dir(self.tree, debian_part,
                file_ids_from=parent_trees + debian_trees)
        rules_path = os.path.join(self.tree.basedir, 'debian', 'rules')
        if os.path.isfile(rules_path):
            os.chmod(rules_path,
                     (stat.S_IRWXU|stat.S_IRGRP|stat.S_IXGRP|
                      stat.S_IROTH|stat.S_IXOTH))
        self.tree.set_parent_ids(parents)
        changelog_path = os.path.join(self.tree.basedir, 'debian',
                'changelog')
        if os.path.exists(changelog_path):
            changelog = self.get_changelog_from_source(
                self.tree.basedir, max_blocks=1)
        message, authors, thanks, bugs = \
                get_commit_info_from_changelog(changelog, self.branch)
        if message is None:
            message = 'Import packaging changes for version %s' % \
                        (str(version),)
        revprops={"deb-md5": md5}
        if native:
            revprops['deb-native'] = "True"
        if authors:
            revprops['authors'] = "\n".join(authors)
        if thanks:
            revprops['deb-thanks'] = "\n".join(thanks)
        if bugs:
            revprops['bugs'] = "\n".join(bugs)
        timezone = None
        if timestamp is not None:
            timezone = timestamp[1]
            timestamp = timestamp[0]
        self._mark_native_config(native)
        revid = self.tree.commit(message, revprops=revprops, timestamp=timestamp,
                timezone=timezone)
        self.tag_version(version, revid=revid)

    def upstream_parents(self, package, versions, version):
        """Get the parents for importing a new upstream.

        The upstream parents will be the last upstream version,
        except for some cases when the last version was native.

        :return: the list of revision ids to use as parents when
            importing the specified upstream version.
        """
        parents = []
        first_parent = self.pristine_upstream_branch.last_revision()
        if first_parent != NULL_REVISION:
            parents = [first_parent]
        last_contained_version = self.last_contained_version(versions)
        if last_contained_version is not None:
            # If the last version was native, and was not from the same
            # upstream as a non-native version (i.e. it wasn't a mistaken
            # native -2 version), then we want to add an extra parent.
            if (self.is_version_native(last_contained_version)
                and not self.pristine_upstream_source.has_version(package,
                    last_contained_version.upstream_version)):
                revid = self.revid_of_version(last_contained_version)
                parents.append(revid)
                self.pristine_upstream_branch.fetch(self.branch,
                        last_revision=revid)
        pull_parents = self.get_parents(versions)
        if ((first_parent == NULL_REVISION and len(pull_parents) > 0)
                or len(pull_parents) > 1):
            if first_parent == NULL_REVISION:
                pull_branch = pull_parents[0][0]
                pull_version = pull_parents[0][1]
            else:
                pull_branch = pull_parents[1][0]
                pull_version = pull_parents[1][1]
            if not pull_branch.is_version_native(pull_version):
                pull_revids = pull_branch.pristine_upstream_source.version_as_revisions(
                    package, pull_version.upstream_version)
                if pull_revids.keys() != [None]:
                    raise MultipleUpstreamTarballsNotSupported()
                pull_revid = pull_revids[None]
                mutter("Initialising upstream from %s, version %s",
                    str(pull_branch), str(pull_version))
                parents.append(pull_revid)
                self.pristine_upstream_branch.fetch(
                        pull_branch.pristine_upstream_branch,
                        last_revision=pull_revid)
                pull_branch.pristine_upstream_branch.tags.merge_to(
                        self.pristine_upstream_branch.tags)
        # FIXME: What about other versions ?
        return { None: parents }

    def get_changelog_from_source(self, dir, max_blocks=None):
        cl_filename = os.path.join(dir, "debian", "changelog")
        content = open(cl_filename).read()
        # Older versions of python-debian were accepting various encodings in
        # the changelog. This is not true with 0.1.20ubuntu2 at least which
        # force an 'utf-8' encoding. This leads to failures when trying to
        # parse old changelogs. Using safe_decode() below will fallback to
        # iso-8859-1 (latin_1) in this case.
        content = safe_decode(content)
        cl = Changelog()
        cl.parse_changelog(content, strict=False, max_blocks=max_blocks)
        return cl

    def _fetch_from_branch(self, branch, revid):
        branch.branch.tags.merge_to(self.branch.tags)
        self.branch.fetch(branch.branch, last_revision=revid)
        if self.pristine_upstream_branch.last_revision() == NULL_REVISION:
            self.pristine_upstream_tree.pull(branch.pristine_upstream_branch)
            branch.pristine_upstream_branch.tags.merge_to(self.pristine_upstream_branch.tags)

    def _import_normal_package(self, package, version, versions, debian_part, md5,
            upstream_part, upstream_tarballs, timestamp=None, author=None,
            file_ids_from=None, pull_debian=True):
        """Import a source package.

        :param package: Package name
        :param version: Full Debian version
        :param versions: Safe versions from changelog
        :param debian_part: Path to extracted directory with Debian changes
        :param unextracted_debian_md5: MD5 sum of unextracted Debian diff/tarball
        :param upstream_part: Extracted upstream directory
        :param upstream_tarballs: List of tuples with (upstream tarfile, md5sum)
        :param timestamp: Version timestamp according to changelog
        :param author: Author according to changelog
        :param file_ids_from: Sequence of trees to take file ids from
        :param pull_debian: Whether to pull from the Debian branch
        """
        pull_branch = None
        if pull_debian:
            pull_branch = self.branch_to_pull_version_from(version, md5)
        if pull_branch is not None:
            if (self.branch_to_pull_upstream_from(package,
                    version.upstream_version, upstream_tarballs)
                    is None):
                pull_branch = None
        if pull_branch is not None:
            self.pull_version_from_branch(pull_branch, package, version)
        else:
            # We need to import at least the diff, possibly upstream.
            # Work out if we need the upstream part first.
            imported_upstream = False
            if not self.pristine_upstream_source.has_version(package,
                    version.upstream_version):
                up_pull_branch = \
                    self.branch_to_pull_upstream_from(package, version.upstream_version,
                            upstream_tarballs)
                if up_pull_branch is not None:
                    self.pull_upstream_from_branch(up_pull_branch,
                            package, version.upstream_version)
                else:
                    imported_upstream = True
                    # Check whether we should pull first if this initialises
                    # from another branch:
                    upstream_parents = self.upstream_parents(package, versions,
                            version.upstream_version)
                    imported_revids = self.import_upstream(upstream_part,
                            package, version.upstream_version,
                            upstream_parents,
                            upstream_tarballs=upstream_tarballs,
                            timestamp=timestamp, author=author,
                            file_ids_from=file_ids_from)
                    self._fetch_upstream_to_branch(imported_revids)
            else:
                mutter("We already have the needed upstream part")
            parents = self.get_parents_with_upstream(package, version, versions,
                    upstream_tarballs, force_upstream_parent=imported_upstream)
            # Now we have the list of parents we need to import the .diff.gz
            self.import_debian(debian_part, version, parents, md5,
                    timestamp=timestamp, file_ids_from=file_ids_from)

    def get_native_parents(self, version, versions):
        last_contained_version = self.last_contained_version(versions)
        if last_contained_version is None:
            parents = []
        else:
            parents = [self.revid_of_version(last_contained_version)]
        missing_versions = self.missing_versions(versions)
        for branch in list(reversed(self.get_lesser_branches())) + self.get_greater_branches():
            merged, missing_versions = \
                branch.contained_versions(missing_versions)
            if merged:
                revid = branch.revid_of_version(merged[0])
                parents.append(revid)
                #FIXME: should this really be here?
                self._fetch_from_branch(branch, revid)
        if (self.branch.last_revision() != NULL_REVISION
                and not self.branch.last_revision() in parents):
            parents.insert(0, self.branch.last_revision())
        return parents

    def _import_native_package(self, package, version, versions, debian_part, md5,
            timestamp=None, file_ids_from=None, pull_debian=True):
        pull_branch = None
        if pull_debian:
            pull_branch = self.branch_to_pull_version_from(version, md5)
        if pull_branch is not None:
            self.pull_version_from_branch(pull_branch, package, version, native=True)
        else:
            parents = self.get_native_parents(version, versions)
            self.import_debian(debian_part, version, parents, md5,
                    native=True, timestamp=timestamp,
                    file_ids_from=file_ids_from)

    def _get_safe_versions_from_changelog(self, cl):
        versions = []
        for block in cl._blocks:
            try:
                versions.append(block.version)
            except VersionError:
                break
        return versions

    def import_package(self, dsc_filename, use_time_from_changelog=True,
            file_ids_from=None, pull_debian=True):
        """Import a source package.

        :param dsc_filename: a path to a .dsc file for the version
            to be imported.
        :param use_time_from_changelog: whether to use the current time or
            the one from the last changelog entry.
        """
        base_path = osutils.dirname(dsc_filename)
        dsc = deb822.Dsc(open(dsc_filename).read())
        version = Version(dsc['Version'])
        format = dsc.get('Format', FORMAT_1_0).strip()
        extractor_cls = SOURCE_EXTRACTORS.get(format)
        if extractor_cls is None:
            raise AssertionError("Don't know how to import source format %s yet"
                    % format)
        extractor = extractor_cls(dsc_filename, dsc)
        try:
            extractor.extract()
            cl = self.get_changelog_from_source(extractor.extracted_debianised)
            timestamp = None
            author = None
            if use_time_from_changelog and len(cl._blocks) > 0:
                 raw_timestamp = cl.date
                 import rfc822, time
                 time_tuple = rfc822.parsedate_tz(raw_timestamp)
                 if time_tuple is not None:
                     timestamp = (calendar.timegm(
                                  time_tuple[:9]) - time_tuple[9],
                                  time_tuple[9])
                 author = safe_decode(cl.author)
            versions = self._get_safe_versions_from_changelog(cl)
            assert not self.has_version(version), \
                "Trying to import version %s again" % str(version)
            #TODO: check that the versions list is correctly ordered,
            # as some methods assume that, and it's not clear what
            # should happen if it isn't.

            if extractor.extracted_upstream is not None:
                self._import_normal_package(dsc['Source'], version, versions,
                        extractor.extracted_debianised,
                        extractor.unextracted_debian_md5,
                        extractor.extracted_upstream,
                        extractor.upstream_tarballs,
                        timestamp=timestamp, author=author,
                        file_ids_from=file_ids_from,
                        pull_debian=pull_debian)
            else:
                self._import_native_package(dsc['Source'], version, versions,
                        extractor.extracted_debianised,
                        extractor.unextracted_debian_md5,
                        timestamp=timestamp, file_ids_from=file_ids_from,
                        pull_debian=pull_debian)
        finally:
            extractor.cleanup()

    def extract_upstream_tree(self, upstream_tips, basedir):
        """Extract upstream_tip to a tempdir as a working tree."""
        # TODO: should stack rather than trying to use the repository,
        # as that will be more efficient.
        to_location = os.path.join(basedir, "upstream")
        # Use upstream_branch if it has been set, otherwise self.branch.
        source_branch = self.pristine_upstream_branch or self.branch
        assert upstream_tips.keys() == [None]
        dir_to = source_branch.bzrdir.sprout(to_location,
                revision_id=upstream_tips[None],
                accelerator_tree=self.tree)
        try:
            self.pristine_upstream_tree = dir_to.open_workingtree()
        except NoWorkingTree:
            # Handle shared treeless repo's.
            self.pristine_upstream_tree = dir_to.create_workingtree()
        self.pristine_upstream_branch = self.pristine_upstream_tree.branch

    def _create_empty_upstream_tree(self, basedir):
        to_location = os.path.join(basedir, "upstream")
        to_transport = get_transport(to_location)
        to_transport.ensure_base()
        format = bzrdir.format_registry.make_bzrdir('default')
        try:
            existing_bzrdir = bzrdir.BzrDir.open_from_transport(
                    to_transport)
        except NotBranchError:
            # really a NotBzrDir error...
            create_branch = bzrdir.BzrDir.create_branch_convenience
            branch = create_branch(to_transport.base,
                    format=format,
                    possible_transports=[to_transport])
        else:
            if existing_bzrdir.has_branch():
                raise AlreadyBranchError(to_location)
            else:
                branch = existing_bzrdir.create_branch()
                existing_bzrdir.create_workingtree()
        self.pristine_upstream_branch = branch
        self.pristine_upstream_tree = branch.bzrdir.open_workingtree()
        if self.tree:
            root_id = self.tree.path2id('')
        else:
            tip = self.branch.basis_tree()
            tip.lock_read()
            try:
                root_id = tip.path2id('')
            finally:
                tip.unlock()
        if root_id:
            self.pristine_upstream_tree.set_root_id(root_id)

    def _extract_tarballs_to_tempdir(self, tarballs):
        tempdir = tempfile.mkdtemp()
        try:
            extract_orig_tarballs(
                [(fn, component) for (fn, component, md5) in tarballs],
                tempdir, strip_components=1)
            return tempdir
        except:
            shutil.rmtree(tempdir)
            raise

    def _export_previous_upstream_tree(self, package, previous_version, tempdir):
        assert isinstance(previous_version, str), \
            "Should pass upstream version as str, not Version."
        try:
            upstream_tips = self.pristine_upstream_source.version_as_revisions(
                    package, previous_version)
        except PackageVersionNotPresent:
            raise BzrCommandError("Unable to find the tag for the "
                    "previous upstream version, %s, in the branch: "
                    "%s" % (
                previous_version,
                self.pristine_upstream_source.tag_name(previous_version)))
        self.extract_upstream_tree(upstream_tips, tempdir)

    def has_merged_upstream_revisions(self, this_revision, upstream_repository, upstream_revisions):
        graph = self.branch.repository.get_graph(
            other_repository=upstream_repository)
        return all(graph.is_ancestor(upstream_revision, this_revision)
                for upstream_revision in upstream_revisions.values())

    def merge_upstream(self, tarball_filenames, package, version, previous_version,
            upstream_branch=None, upstream_revisions=None, merge_type=None,
            force=False):
        assert isinstance(version, str), \
            "Should pass version as str not %s" % str(type(version))
        assert isinstance(previous_version, str) or previous_version is None, \
            "Should pass previous_version as str not %s" % str(
                    type(previous_version))
        tempdir = tempfile.mkdtemp(dir=os.path.join(self.tree.basedir, '..'))
        try:
            if previous_version is not None:
                self._export_previous_upstream_tree(package, previous_version, tempdir)
            else:
                self._create_empty_upstream_tree(tempdir)
            if self.pristine_upstream_source.has_version(package, version):
                raise UpstreamAlreadyImported(version)
            if upstream_branch is not None:
                upstream_branch.lock_read()
            try:
                if upstream_branch is not None:
                    if upstream_revisions is None:
                        upstream_revisions = { None: upstream_branch.last_revision() }
                    if (not force and
                        self.has_merged_upstream_revisions(self.branch.last_revision(), upstream_branch.repository, upstream_revisions)):
                        raise UpstreamBranchAlreadyMerged
                upstream_tarballs = [
                    (os.path.abspath(fn), component, md5sum_filename(fn)) for
                    (fn, component) in
                    tarball_filenames]
                tarball_dir = self._extract_tarballs_to_tempdir(upstream_tarballs)
                try:
                    # FIXME: should use upstream_parents()?
                    parents = { None: [] }
                    if self.pristine_upstream_branch.last_revision() != NULL_REVISION:
                        parents = { None: [self.pristine_upstream_branch.last_revision()] }
                    imported_revids = self.import_upstream(tarball_dir,
                            package, version, parents,
                            upstream_tarballs=upstream_tarballs,
                            upstream_branch=upstream_branch,
                            upstream_revisions=upstream_revisions)
                    self._fetch_upstream_to_branch(imported_revids)
                finally:
                    shutil.rmtree(tarball_dir)
                if self.branch.last_revision() != NULL_REVISION:
                    try:
                        conflicts = self.tree.merge_from_branch(
                                self.pristine_upstream_branch, merge_type=merge_type)
                    except UnrelatedBranches:
                        # Bug lp:515367 where the first upstream tarball is
                        # missing a proper history link and a criss-cross merge
                        # then recurses and finds no deeper ancestor.
                        # Use the previous upstream import as the from revision
                        if len(parents[None]) == 0:
                            from_revision = NULL_REVISION
                        else:
                            from_revision = parents[None][0]
                        conflicts = self.tree.merge_from_branch(
                                self.pristine_upstream_branch, merge_type=merge_type,
                                from_revision=from_revision)
                else:
                    # Pull so that merge-upstream allows you to start a branch
                    # from upstream tarball.
                    conflicts = 0
                    self.tree.pull(self.pristine_upstream_branch)
                self.pristine_upstream_branch.tags.merge_to(self.branch.tags)
                return conflicts
            finally:
                if upstream_branch is not None:
                    upstream_branch.unlock()
        finally:
            shutil.rmtree(tempdir)


class SourceExtractor(object):
    """A class to extract a source package to its constituent parts"""

    def __init__(self, dsc_path, dsc):
        self.dsc_path = dsc_path
        self.dsc = dsc
        self.extracted_upstream = None
        self.extracted_debianised = None
        self.unextracted_debian_md5 = None
        self.upstream_tarballs = []
        self.tempdir = None

    def extract(self):
        """Extract the package to a new temporary directory."""
        raise NotImplementedError(self.extract)

    def cleanup(self):
        """Cleanup any extracted files."""
        if self.tempdir is not None and os.path.isdir(self.tempdir):
            shutil.rmtree(self.tempdir)


class OneZeroSourceExtractor(SourceExtractor):
    """Source extract for the "1.0" source format."""

    def extract(self):
        """Extract the package to a new temporary directory."""
        self.tempdir = tempfile.mkdtemp()
        dsc_filename = os.path.abspath(self.dsc_path)
        proc = subprocess.Popen("dpkg-source -su -x %s" % (dsc_filename,), shell=True,
                cwd=self.tempdir, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT, preexec_fn=subprocess_setup)
        (stdout, _) = proc.communicate()
        assert proc.returncode == 0, "dpkg-source -x failed, output:\n%s" % \
                    (stdout,)
        name = self.dsc['Source']
        version = Version(self.dsc['Version'])
        self.extracted_upstream = os.path.join(self.tempdir,
                "%s-%s.orig" % (name, str(version.upstream_version)))
        self.extracted_debianised = os.path.join(self.tempdir,
                "%s-%s" % (name, str(version.upstream_version)))
        if not os.path.exists(self.extracted_upstream):
            mutter("It's a native package")
            self.extracted_upstream = None
        for part in self.dsc['files']:
            if self.extracted_upstream is None:
                if part['name'].endswith(".tar.gz"):
                    self.unextracted_debian_md5 = part['md5sum']
            else:
                if part['name'].endswith(".orig.tar.gz"):
                    self.upstream_tarballs.append((os.path.abspath(
                            os.path.join(osutils.dirname(self.dsc_path),
                                part['name'])),
                            component_from_orig_tarball(part['name'], name, str(version.upstream_version)),
                            part['md5sum']))
                elif part['name'].endswith(".diff.gz"):
                    self.unextracted_debian_md5 = part['md5sum']


class ThreeDotZeroNativeSourceExtractor(SourceExtractor):
    """Source extractor for the "3.0 (native)" source format."""

    def extract(self):
        self.tempdir = tempfile.mkdtemp()
        dsc_filename = os.path.abspath(self.dsc_path)
        proc = subprocess.Popen("dpkg-source -x %s" % (dsc_filename,), shell=True,
                cwd=self.tempdir, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT, preexec_fn=subprocess_setup)
        (stdout, _) = proc.communicate()
        assert proc.returncode == 0, "dpkg-source -x failed, output:\n%s" % \
                    (stdout,)
        name = self.dsc['Source']
        version = Version(self.dsc['Version'])
        self.extracted_debianised = os.path.join(self.tempdir,
                "%s-%s" % (name, str(version.upstream_version)))
        self.extracted_upstream = None
        for part in self.dsc['files']:
            if (part['name'].endswith(".tar.gz")
                    or part['name'].endswith(".tar.bz2")
                    or part['name'].endswith(".tar.xz")):
                self.unextracted_debian_md5 = part['md5sum']


class ThreeDotZeroQuiltSourceExtractor(SourceExtractor):
    """Source extractor for the "3.0 (quilt)" source format."""

    def extract(self):
        self.tempdir = tempfile.mkdtemp()
        dsc_filename = os.path.abspath(self.dsc_path)
        proc = subprocess.Popen("dpkg-source --skip-debianization -x %s"
                % (dsc_filename,), shell=True,
                cwd=self.tempdir, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT, preexec_fn=subprocess_setup)
        (stdout, _) = proc.communicate()
        assert proc.returncode == 0, "dpkg-source -x failed, output:\n%s" % \
                    (stdout,)
        name = self.dsc['Source']
        version = Version(self.dsc['Version'])
        self.extracted_debianised = os.path.join(self.tempdir,
                "%s-%s" % (name, str(version.upstream_version)))
        self.extracted_upstream = self.extracted_debianised + ".orig"
        os.rename(self.extracted_debianised, self.extracted_upstream)
        proc = subprocess.Popen("dpkg-source -x %s" % (dsc_filename,), shell=True,
                cwd=self.tempdir, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT, preexec_fn=subprocess_setup)
        (stdout, _) = proc.communicate()
        assert proc.returncode == 0, "dpkg-source -x failed, output:\n%s" % \
                    (stdout,)
        # Check that there are no unreadable files extracted.
        subprocess.call(["find", self.extracted_upstream, "-perm",
                "0000", "-exec", "chmod", "644", "{}", ";"])
        subprocess.call(["find", self.extracted_debianised, "-perm",
                "0000", "-exec", "chmod", "644", "{}", ";"])
        for part in self.dsc['files']:
            if part['name'].startswith("%s_%s.orig" % (name, str(version.upstream_version))):
                self.upstream_tarballs.append((
                    os.path.abspath(os.path.join(osutils.dirname(self.dsc_path),
                                    part['name'])),
                    component_from_orig_tarball(part['name'], name,
                        str(version.upstream_version)),
                    part['md5sum']))
            elif (part['name'].endswith(".debian.tar.gz")
                    or part['name'].endswith(".debian.tar.bz2")
                    or part['name'].endswith(".debian.tar.xz")):
                self.unextracted_debian_md5 = part['md5sum']
        assert self.upstream_tarballs is not None, \
            "Can't handle non gz|bz2|xz tarballs yet"
        assert self.unextracted_debian_md5 is not None, \
            "Can't handle non gz|bz2|xz tarballs yet"


SOURCE_EXTRACTORS = {}
SOURCE_EXTRACTORS[FORMAT_1_0] = OneZeroSourceExtractor
SOURCE_EXTRACTORS[FORMAT_3_0_NATIVE] = ThreeDotZeroNativeSourceExtractor
SOURCE_EXTRACTORS[FORMAT_3_0_QUILT] = ThreeDotZeroQuiltSourceExtractor



#    quilt.py -- Quilt patch handling
#    Copyright (C) 2011 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""Quilt patch handling."""

from __future__ import absolute_import

import errno
import os
import signal
import subprocess
from bzrlib import (
    errors,
    trace,
    )


class QuiltError(errors.BzrError):

    _fmt = "An error (%(retcode)d) occurred running quilt: %(stderr)s%(extra)s"

    def __init__(self, retcode, stdout, stderr):
        self.retcode = retcode
        self.stderr = stderr
        if stdout is not None:
            self.extra = "\n\n%s" % stdout
        else:
            self.extra = ""
        self.stdout = stdout


def run_quilt(args, working_dir, series_file=None, patches_dir=None,
        quiet=None):
    """Run quilt.

    :param args: Arguments to quilt
    :param working_dir: Working dir
    :param series_file: Optional path to the series file
    :param patches_dir: Optional path to the patches
    :param quilt: Whether to be quiet (quilt stderr not to terminal)
    :raise QuiltError: When running quilt fails
    """
    def subprocess_setup():
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    env = {}
    if patches_dir is not None:
        env["QUILT_PATCHES"] = patches_dir
    else:
        env["QUILT_PATCHES"] = os.path.join(working_dir, "debian", "patches")
    if series_file is not None:
        env["QUILT_SERIES"] = series_file
    else:
        env["QUILT_SERIES"] = os.path.join(env["QUILT_PATCHES"], "series")
    # Hide output if -q is in use.
    if quiet is None:
        quiet = trace.is_quiet()
    if not quiet:
        stderr =  subprocess.STDOUT
    else:
        stderr = subprocess.PIPE
    command = ["quilt"] + args
    trace.mutter("running: %r", command)
    if not os.path.isdir(working_dir):
        raise AssertionError("%s is not a valid directory" % working_dir)
    try:
        proc = subprocess.Popen(command, cwd=working_dir, env=env,
                stdin=subprocess.PIPE, preexec_fn=subprocess_setup,
                stdout=subprocess.PIPE, stderr=stderr)
    except OSError, e:
        if e.errno != errno.ENOENT:
            raise
        raise errors.BzrError("quilt is not installed, please install it")
    output = proc.communicate()
    if proc.returncode not in (0, 2):
        raise QuiltError(proc.returncode, output[0], output[1])
    if output[0] is None:
        return ""
    return output[0]


def quilt_pop_all(working_dir, patches_dir=None, series_file=None, quiet=None,
        force=False):
    """Pop all patches.

    :param working_dir: Directory to work in
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    args = ["pop", "-a"]
    if force:
        args.append("-f")
    return run_quilt(args, working_dir=working_dir,
        patches_dir=patches_dir, series_file=series_file, quiet=quiet)


def quilt_pop(working_dir, patch, patches_dir=None, series_file=None, quiet=None):
    """Pop a patch.

    :param working_dir: Directory to work in
    :param patch: Patch to apply
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    return run_quilt(["pop", patch], working_dir=working_dir,
        patches_dir=patches_dir, series_file=series_file, quiet=quiet)


def quilt_push_all(working_dir, patches_dir=None, series_file=None, quiet=None,
        force=False):
    """Push all patches.

    :param working_dir: Directory to work in
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    args = ["push", "-a"]
    if force:
        args.append("-f")
    return run_quilt(args, working_dir=working_dir,
            patches_dir=patches_dir, series_file=series_file, quiet=quiet)


def quilt_push(working_dir, patch, patches_dir=None, series_file=None, quiet=None):
    """Push a patch.

    :param working_dir: Directory to work in
    :param patch: Patch to push
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    return run_quilt(["push", patch], working_dir=working_dir,
            patches_dir=patches_dir, series_file=series_file, quiet=quiet)


def quilt_applied(tree):
    """Find the list of applied quilt patches.

    """
    file_id = tree.path2id(".pc/applied-patches")
    if file_id is None:
        return []
    try:
        return [patch.rstrip("\n") for patch in
            tree.get_file_lines(file_id, ".pc/applied-patches")
            if patch.strip() != ""]
    except (IOError, OSError), e:
        if e.errno == errno.ENOENT:
            # File has already been removed
            return []
        raise


def quilt_unapplied(working_dir, patches_dir=None, series_file=None):
    """Find the list of unapplied quilt patches.

    :param working_dir: Directory to work in
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    try:
        unapplied_patches = run_quilt(["unapplied"],
                            working_dir=working_dir, patches_dir=patches_dir,
                            series_file=series_file).splitlines()
        patch_basenames = []
        for patch in unapplied_patches:
            patch = os.path.basename(patch)
            patch_basenames.append(patch)
        return patch_basenames
    except QuiltError, e:
        if e.retcode == 1:
            return []
        raise


def quilt_series(tree):
    """Find the list of patches.

    :param tree: Tree to read from
    """
    file_id = tree.path2id("debian/patches/series")
    if file_id is None:
        return []
    try:
        return [patch.rstrip("\n") for patch in
            tree.get_file_lines(file_id, "debian/patches/series")
            if patch.strip() != ""]
    except (IOError, OSError), e:
        if e.errno == errno.ENOENT:
            # File has already been removed
            return []
        raise



# Copyright (C) 2010 Canonical Limited
# vim: ts=4 sts=4 sw=4
#
# This file is part of bzr-builddeb.
#
# bzr-builddeb is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# bzr-builddeb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bzr-builddeb; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""Tests for bzrlib.plugins.builddeb.tagging."""

from bzrlib.plugins.builddeb import tagging
from bzrlib.tests import (
    TestCase,
    )


class TestDebVersionSort(TestCase):

    def test_sort(self):
        tags = [("1.0", "revid"), ("1.0.1", "revid"), ("1.0~1", "revid")]
        tagging.sort_debversion(None, tags)
        self.assertEquals(tags,
            [("1.0~1", "revid"), ("1.0", "revid"), ("1.0.1", "revid")])



#    test_mark_uploaded.py -- Blackbox tests for mark-uploaded.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

try:
    from debian.changelog import Changelog, Version
except ImportError:
    # Prior to 0.1.15 the debian module was called debian_bundle
    from debian_bundle.changelog import Changelog, Version

from bzrlib.plugins.builddeb.tests import BuilddebTestCase


class TestMarkUploaded(BuilddebTestCase):

    def make_unuploaded(self):
        self.wt = self.make_branch_and_tree('.')
        self.build_tree(['debian/'])
        cl = Changelog()
        v = Version("0.1-1")
        cl.new_block(package='package',
                     version=Version('0.1-1'),
                     distributions='unstable',
                     urgency='low',
                     author='James Westby <jw+debian@jameswestby.net>',
                     date='Thu,  3 Aug 2006 19:16:22 +0100',
                     )
        cl.add_change('');
        cl.add_change('  * Initial packaging.');
        cl.add_change('');
        f = open('debian/changelog', 'wb')
        try:
            cl.write_to_open_file(f)
        finally:
            f.close()
        self.wt.add(["debian/", "debian/changelog"])
        self.wt.commit("one")

    def test_mark_uploaded_available(self):
        self.run_bzr('mark-uploaded --help')

    def test_mark_uploaded_changes(self):
        self.make_unuploaded()
        self.build_tree(['foo'])
        self.wt.add(['foo'])
        self.run_bzr_error(["There are uncommitted changes"],
                "mark-uploaded")

    def test_mark_uploaded_unkown_dist(self):
        self.make_unuploaded()
        cl = Changelog()
        v = Version("0.1-1")
        cl.new_block(package='package',
                     version=Version('0.1-1'),
                     distributions='UNRELEASED',
                     urgency='low',
                     author='James Westby <jw+debian@jameswestby.net>',
                     date='Thu,  3 Aug 2006 19:16:22 +0100',
                     )
        cl.add_change('');
        cl.add_change('  * Initial packaging.');
        cl.add_change('');
        f = open('debian/changelog', 'wb')
        try:
            cl.write_to_open_file(f)
        finally:
            f.close()
        self.wt.commit("two")
        self.run_bzr_error(["The changelog still targets 'UNRELEASED', so "
                "apparently hasn't been uploaded."], "mark-uploaded")

    def test_mark_uploaded_already(self):
        self.make_unuploaded()
        self.run_bzr("mark-uploaded")
        self.build_tree(["foo"])
        self.wt.add(["foo"])
        self.wt.commit("two")
        self.run_bzr_error(["This version has already been marked uploaded"],
                "mark-uploaded")

    def test_mark_uploaded(self):
        self.make_unuploaded()
        self.run_bzr("mark-uploaded")
        tagged_revision = self.wt.branch.tags.lookup_tag('0.1-1')
        self.assertEqual(tagged_revision, self.wt.branch.last_revision())

    def test_mark_uploaded_force(self):
        self.make_unuploaded()
        self.run_bzr("mark-uploaded")
        self.build_tree(["foo"])
        self.wt.add(["foo"])
        self.wt.commit("two")
        self.run_bzr("mark-uploaded --force")
        tagged_revision = self.wt.branch.tags.lookup_tag('0.1-1')
        self.assertEqual(tagged_revision, self.wt.branch.last_revision())



#    launchpad.py -- Lookups via. launchpadlib
#    Copyright (C) 2009 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os

from bzrlib.config import config_dir
from bzrlib.trace import (
    mutter,
    note,
    )

try:
    from launchpadlib.launchpad import Launchpad
    from launchpadlib.credentials import Credentials
    from launchpadlib.uris import LPNET_SERVICE_ROOT
    HAVE_LPLIB = True
except ImportError:
    HAVE_LPLIB = False


def get_launchpad():
    if not HAVE_LPLIB:
        return None
    try:
        return Launchpad.login_anonymously("bzr-builddeb",
            service_root=LPNET_SERVICE_ROOT)
    except AttributeError: # older version of launchpadlib
        creds_path = os.path.join(config_dir(), "builddeb.lp_creds.txt")
        if not os.path.exists(creds_path):
            return None
        creds = Credentials("bzr-builddeb")
        f = open(creds_path)
        try:
            creds.load(f)
        finally:
            f.close()
        return Launchpad(creds, service_root=LPNET_SERVICE_ROOT)


def ubuntu_bugs_for_debian_bug(bug_id):
    """Find Ubuntu bugs linked to a particular Debian bug.

    :param bug_id: Debian bug id
    :return: Liust of Launchpad bug ids for Ubuntu bugs
    """
    lp = get_launchpad()
    if lp is None:
        return []
    try:
        bug = lp.load(str(lp._root_uri) + "bugs/bugtrackers/debbugs/%s")
        tasks = bug.bug_tasks
        for task in tasks:
            if task.bug_target_name.endswith("(Ubuntu)"):
                return str(bug.id)
    except Exception, e:
        mutter(str(e))
        return []
    return []


def debian_bugs_for_ubuntu_bug(bug_id):
    """Find the Debian bugs linked to a particular Ubuntu bug.

    :param bug_id: The Launchpad bug ID for the Ubuntu bug
    :return: List of Debian bug URLs.
    """
    lp = get_launchpad()
    if lp is None:
        return []
    try:
        bug = lp.bugs[int(bug_id)]
        tasks = bug.bug_tasks
        for task in tasks:
            if task.bug_target_name.endswith("(Debian)"):
                watch = task.bug_watch
                if watch is None:
                    break
                return watch.remote_bug.encode("utf-8")
    except Exception, e:
        mutter(str(e))
        return []
    return []


def get_upstream_projectseries_for_package(package, distribution_name, distroseries_name):
    """Return the upstream project series for a package.

    :param package: Package name
    :param distribution_name: Distribution name
    :param distroseries_name: Distribution series name
    :return: Launchpad object for the upstream project series
    """
    lp = get_launchpad()
    if lp is None:
        return None
    distribution = lp.distributions[distribution_name]
    if distribution is None:
        note("Launchpad: No such distribution %s" % distribution)
        return None
    distroseries = distribution.getSeries(name_or_version=distroseries_name)
    if distroseries is None:
        note("%s: No such distroseries %s" % (distribution_name, distroseries_name))
        return None
    sourcepackage = distroseries.getSourcePackage(name=package)
    if sourcepackage is None:
        note("%s: Source package %s not found in %s" % (distribution_name, package,
            sourcepackage))
        return None
    productseries = sourcepackage.productseries
    if productseries is None:
        note("%s: Source package %s in %s not linked to a product series" % (
            distribution_name, package, sourcepackage))
        return None
    return productseries


def get_upstream_branch_url(package, distribution_name, distroseries_name):
    """Return the upstream branch URL based on a package in a distribution.

    :param distribution_name: Distribution name
    :param package: Source package name
    :param distroseries_name: Distroseries name
    """
    projectseries = get_upstream_projectseries_for_package(
        package, distribution_name, distroseries_name)
    if projectseries is None:
        return
    branch = projectseries.branch
    if branch is None:
        note(("%s: upstream project series %s for source package %s does "
              "not have a branch") % (distribution_name, distroseries_name,
                  package))
        return None
    return branch.bzr_identity



#    test_builddeb.py -- Blackbox tests for builddeb.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#    
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os

from bzrlib.plugins.builddeb.tests import BuilddebTestCase


class TestBuilddeb(BuilddebTestCase):

  commited_file = 'commited_file'
  uncommited_file = 'uncommited_file'
  unadded_file = 'unadded_file'

  def make_unpacked_source(self):
    """Create an unpacked source tree in a branch. Return the working tree"""
    tree = self.make_branch_and_tree('.')
    cl_file = 'debian/changelog'
    source_files = ['README', 'debian/'] + [cl_file]
    self.build_tree(source_files)
    c = self.make_changelog()
    self.write_changelog(c, cl_file)
    tree.add(source_files)
    return tree

  def build_dir(self):
    return os.path.join('..', 'build-area',
                        self.package_name + '-' +
                        str(self.upstream_version))

  def assertInBuildDir(self, files):
    build_dir = self.build_dir()
    if isinstance(files, basestring):
      files = [files]
    for filename in files:
      self.assertPathExists(os.path.join(build_dir, filename))

  def assertNotInBuildDir(self, files):
    build_dir = self.build_dir()
    if isinstance(files, basestring):
      files = [files]
    for filename in files:
      self.assertPathDoesNotExist(os.path.join(build_dir, filename))

  def test_builddeb_registered(self):
    self.run_bzr("builddeb --help")

  def test_bd_alias(self):
    self.run_bzr("bd --help")

  def test_builddeb_not_package(self):
    self.run_bzr_error(['Could not find changelog'], 'builddeb')

  def build_really_simple_tree(self):
    tree = self.make_unpacked_source()
    self.build_tree([self.commited_file, self.uncommited_file,
                     self.unadded_file])
    tree.add([self.commited_file])
    tree.commit("one", rev_id='revid1')
    tree.add([self.uncommited_file])
    return tree

  def build_tree_with_conflict(self):
    tree = self.make_unpacked_source()
    self.build_tree([self.commited_file, self.uncommited_file,
                     self.unadded_file])
    tree.add([self.commited_file])
    tree.commit("one", rev_id='revid1')
    newtree = tree.bzrdir.sprout('newtree').open_workingtree()
    tree.add([self.uncommited_file])
    tree.commit("two", rev_id='revid2')

    p = '%s/work/newtree/%s' % (self.test_base_dir, self.uncommited_file)
    fh = open(p, 'w')
    fh.write('** This is the conflicting line.')
    fh.close()
    newtree.add([self.uncommited_file])
    newtree.commit("new-two", rev_id='revidn2')
    conflicts = tree.merge_from_branch(newtree.branch) 

    return (conflicts, tree)

  def test_builddeb_uses_working_tree(self):
    self.build_really_simple_tree()
    self.run_bzr("builddeb --native --builder true --dont-purge")
    self.assertInBuildDir([self.commited_file, self.uncommited_file])
    self.assertNotInBuildDir([self.unadded_file])

  def test_builddeb_refuses_tree_with_conflicts(self):
    (conflicts, tree) = self.build_tree_with_conflict()
    self.assertTrue(conflicts > 0)
    self.run_bzr_error(
      ['There are conflicts in the working tree. You must resolve these'],
      "builddeb --native --builder true --dont-purge")

  def test_builddeb_strict(self):
    tree = self.build_really_simple_tree()
    self.run_bzr_error(
      ['bzr: ERROR: Build refused because there are unknown files in the tree. '
       "To list all known files, run 'bzr unknowns'."],
      "builddeb --strict")

  def test_builddeb_uses_revision_when_told(self):
    self.build_really_simple_tree()
    self.run_bzr("builddeb --native --builder true --dont-purge -r-1")
    self.assertInBuildDir([self.commited_file])
    self.assertNotInBuildDir([self.unadded_file, self.uncommited_file])

  def test_builddeb_error_on_two_revisions(self):
    tree = self.make_unpacked_source()
    self.run_bzr_error(['--revision takes exactly one revision specifier.'],
                       "builddeb --native --builder true -r0..1")

  def test_builddeb_allows_building_revision_0(self):
    self.build_really_simple_tree()
    # This may break if there is something else that needs files in the
    # branch before the changelog is looked for.
    self.run_bzr_error(['Could not find changelog'],
                       "builddeb --native --builder true --dont-purge -r0")
    self.assertNotInBuildDir([self.commited_file, self.unadded_file,
                              self.uncommited_file])

  def orig_dir(self):
    return '..'

  def make_upstream_tarball(self):
    f = open(os.path.join(self.orig_dir(), self.package_name + "_" +
                          str(self.package_version.upstream_version) +
                          ".orig.tar.gz"), 'wb')
    f.close()

  def test_builder(self):
    tree = self.make_unpacked_source()
    self.run_bzr('bd --dont-purge --native --builder "touch built"')
    self.assertInBuildDir('built')

  def test_package_merge_first(self):
    tree = self.make_unpacked_source()
    self.make_upstream_tarball()
    self.run_bzr('bd --package-merge --builder true')

  def test_hooks(self):
    tree = self.make_unpacked_source()
    self.make_upstream_tarball()
    os.mkdir('.bzr-builddeb/')
    f = open('.bzr-builddeb/default.conf', 'wb')
    try:
      f.write('[HOOKS]\npre-export = touch pre-export\n')
      f.write('pre-build = touch pre-build\npost-build = touch post-build\n')
    finally:
      f.close()
    self.run_bzr('add .bzr-builddeb/default.conf')
    self.run_bzr('bd --dont-purge --builder true')
    self.assertPathExists('pre-export')
    self.assertInBuildDir(['pre-build', 'post-build'])

  def test_utf8_changelog(self):
    from bzrlib.plugins import builddeb
    from bzrlib.msgeditor import hooks
    if "set_commit_message" not in hooks:
      self.skip("No set_commit_message hook in bzrlib this old")
    hooks.install_named_hook("set_commit_message",
      builddeb.debian_changelog_commit, "Test builddeb set commit msg hook")
    tree = self.make_unpacked_source()
    f = open("debian/bzr-builddeb.conf", 'wb')
    try:
      f.write("[BUILDDEB]\ncommit-message-from-changelog = true")
    finally:
      f.close()
    tree.add("debian/bzr-builddeb.conf")
    # The changelog is only used for commit message when it already exists and
    # is then changed, so need to clear it, commit, then set the contents.
    open("debian/changelog", "w").close()
    tree.commit("Prepare for release", rev_id="prerel")
    # Would be nice to use debian.changelog to make one, but it's changed
    # across versions as to how it wants non-ascii bytes provided.
    f = open("debian/changelog", "w")
    try:
      f.write("  * \xe2\x80\xa6and another thing")
    finally:
      f.close()
    self.run_bzr(['commit'])
    branch = tree.branch
    self.assertEqual(u"\u2026and another thing",
      branch.repository.get_revision(branch.last_revision()).message)

# vim: ts=2 sts=2 sw=2



#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-
#    test_merge_package.py -- Merge packaging branches, fix ancestry as needed.
#    Copyright (C) 2009 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import string
import unittest

try:
    from debian.changelog import Version
except ImportError:
    # Prior to 0.1.15 the debian module was called debian_bundle
    from debian_bundle.changelog import Version

from bzrlib.plugins.builddeb import merge_package as MP
from bzrlib.plugins.builddeb.errors import (
    SharedUpstreamConflictsWithTargetPackaging)
from bzrlib.plugins.builddeb.import_dsc import DistributionBranch

from bzrlib.plugins.builddeb.tests import TestCaseWithTransport

_Debian_changelog = '''\
ipsec-tools (%s) unstable; urgency=high

  * debian packaging -- %s

 -- Nico Golde <nion@debian.org>  Tue, %02d May 2009 13:26:14 +0200

'''

_Ubuntu_changelog = '''\
ipsec-tools (%s) karmic; urgency=low

  * ubuntu packaging -- %s

 -- Jamie Strandboge <jamie@ubuntu.com>  Fri, %02d Jul 2009 13:24:17 -0500

'''


def _prepend_log(text, path):
    content = open(path).read()
    fh = open(path, 'wb')
    try:
        fh.write(text+content)
    finally:
        fh.close()


class MergePackageTests(TestCaseWithTransport):

    def test__upstream_version_data(self):
        ubup_o, debp_n, _ubuu, _debu = self._setup_debian_upstream_newer()
        vdata = MP._upstream_version_data(debp_n.branch, debp_n.last_revision())
        self.assertEquals(vdata[0], Version('1.10'))

        vdata = MP._upstream_version_data(ubup_o.branch, ubup_o.last_revision())
        self.assertEquals(vdata[0], Version('1.2'))

    def test_debian_upstream_newer(self):
        """Diverging upstreams (debian newer) don't cause merge conflicts.

        The debian and ubuntu upstream branches will differ with regard to
        the content of the file 'c'.

        Furthermore the respective packaging branches will have a text
        conflict in 'debian/changelog'.

        The upstream conflict will be resolved by fix_ancestry_as_needed().
        Please note that the debian ancestry is more recent.
        """
        ubup, debp, ubuu, debu = self._setup_debian_upstream_newer()

        # Attempt a plain merge first.
        conflicts = ubup.merge_from_branch(
            debp.branch, to_revision=self.revid_debp_n_C)

        # There are two conflicts in the 'c' and the 'debian/changelog' files
        # respectively.
        self.assertEquals(conflicts, 2)
        conflict_paths = sorted([c.path for c in ubup.conflicts()])
        self.assertEquals(conflict_paths, [u'c.moved', u'debian/changelog'])

        # Undo the failed merge.
        ubup.revert()

        # Check the versions present in the tree with the fixed ancestry.
        v3 = "1.2"
        v4 = "1.10"
        db1 = DistributionBranch(ubup.branch, ubup.branch)
        self.assertEqual(db1.pristine_upstream_source.has_version("package", v3), True)
        # This version is in the diverged debian upstream tree and will
        # hence not be present in the target ubuntu packaging branch.
        self.assertEqual(db1.pristine_upstream_source.has_version("package", v4), False)

        # The ubuntu upstream branch tip.
        ubuu_tip = ubuu.branch.last_revision()
        # The debian upstream branch tip.
        debu_tip = debu.branch.last_revision()
        # The ubuntu packaging branch tip.
        ubup_tip_pre_fix = ubup.branch.last_revision()

        # The first conflict is resolved by calling fix_ancestry_as_needed().
        upstreams_diverged, t_upstream_reverted = MP.fix_ancestry_as_needed(ubup, debp.branch)

        # The ancestry did diverge and needed to be fixed.
        self.assertEquals(upstreams_diverged, True)
        # The (temporary) target upstream branch had to be reverted to the
        # source upstream branch since the latter was more recent.
        self.assertEquals(t_upstream_reverted, True)

        # Check the versions present in the tree with the fixed ancestry.
        db2 = DistributionBranch(ubup.branch, ubup.branch)
        self.assertEqual(db2.pristine_upstream_source.has_version("package", v3), True)
        # The ancestry has been fixed and the missing debian upstream
        # version should now be present in the target ubuntu packaging
        # branch.
        self.assertEqual(db2.pristine_upstream_source.has_version("package", v4), True)

        # Now let's take a look at the fixed ubuntu packaging branch.
        ubup_tip_post_fix = ubup.branch.last_revision()
        ubup_parents_post_fix = ubup.branch.repository.revision_tree(ubup_tip_post_fix).get_parent_ids()

        # The tip of the fixed ubuntu packaging branch has 2 parents.
        self.assertEquals(len(ubup_parents_post_fix), 2)

        # The left parent is the packaging branch tip before fixing.
        self.assertEquals(ubup_parents_post_fix[0], ubup_tip_pre_fix)

        # The right parent is derived from a merge 
        ubup_parents_sharedupstream = ubup.branch.repository.revision_tree(ubup_parents_post_fix[1]).get_parent_ids()
        self.assertEquals(ubup_parents_sharedupstream, [ubuu_tip, debu_tip])

        # Try merging again.
        conflicts = ubup.merge_from_branch(
            debp.branch, to_revision=self.revid_debp_n_C)

        # And, voila, only the packaging branch conflict remains.
        self.assertEquals(conflicts, 1)
        conflict_paths = sorted([c.path for c in ubup.conflicts()])
        self.assertEquals(conflict_paths, [u'debian/changelog'])

    def test_deb_upstream_conflicts_with_ubu_packaging(self):
        """Source upstream conflicts with target packaging -> exception.

        The debian upstream and the ubuntu packaging branches will differ
        with respect to the content of the file 'c'.

        The conflict cannot be resolved by fix_ancestry_as_needed().
        The `SharedUpstreamConflictsWithTargetPackaging` exception is
        thrown instead.
        """
        ubup, debp, ubuu, debu = self._setup_debian_upstream_conflicts()

        e = self.assertRaises(
            SharedUpstreamConflictsWithTargetPackaging,
            MP.fix_ancestry_as_needed, ubup, debp.branch)

        conflict_paths = sorted([c.path for c in ubup.conflicts()])
        self.assertEquals(conflict_paths, [u'c.moved'])
        # Check that all the merged revisions are now in this repo
        merged_parent = ubup.get_parent_ids()[1]
        its_parents_map = ubup.branch.repository.get_parent_map([
                merged_parent])
        self.assertTrue(merged_parent in its_parents_map)
        its_parents = its_parents_map[merged_parent]
        their_parents = ubup.branch.repository.get_parent_map(its_parents)
        self.assertTrue(its_parents[0] in their_parents)
        self.assertTrue(its_parents[1] in their_parents)

    def test_debian_upstream_older(self):
        """Diverging upstreams (debian older) don't cause merge conflicts.

        The debian and ubuntu upstream branches will differ with regard to
        the content of the file 'c'.

        Furthermore the respective packaging branches will have a text
        conflict in 'debian/changelog'.

        The upstream conflict will be resolved by fix_ancestry_as_needed().
        Please note that the debian ancestry is older in this case.
        """
        ubup, debp, _ubuu, _debu = self._setup_debian_upstream_older()

        # Attempt a plain merge first.
        conflicts = ubup.merge_from_branch(
            debp.branch, to_revision=self.revid_debp_o_C)

        # There are two conflicts in the 'c' and the 'debian/changelog' files
        # respectively.
        self.assertEquals(conflicts, 2)
        conflict_paths = sorted([c.path for c in ubup.conflicts()])
        self.assertEquals(conflict_paths, [u'c.moved', u'debian/changelog'])

        # Undo the failed merge.
        ubup.revert()

        # The first conflict is resolved by calling fix_ancestry_as_needed().
        upstreams_diverged, t_upstream_reverted = MP.fix_ancestry_as_needed(ubup, debp.branch)

        # The ancestry did diverge and needed to be fixed.
        self.assertEquals(upstreams_diverged, True)
        # The target upstream branch was more recent in this case and hence
        # was not reverted to the source upstream branch.
        self.assertEquals(t_upstream_reverted, False)

        # Try merging again.
        conflicts = ubup.merge_from_branch(
            debp.branch, to_revision=self.revid_debp_o_C)

        # And, voila, only the packaging branch conflict remains.
        self.assertEquals(conflicts, 1)
        conflict_paths = sorted([c.path for c in ubup.conflicts()])
        self.assertEquals(conflict_paths, [u'debian/changelog'])

    def test_upstreams_not_diverged(self):
        """Non-diverging upstreams result in a normal merge.

        The debian and ubuntu upstream branches will not have diverged
        this time.

        The packaging branches will have a conflict in 'debian/changelog'.
        fix_ancestry_as_needed() will return as soon as establishing that
        the upstreams have not diverged.
        """
        ubuntup, debianp = self._setup_upstreams_not_diverged()

        # Attempt a plain merge first.
        conflicts = ubuntup.merge_from_branch(
            debianp.branch, to_revision=self.revid_debianp_C)

        # There is only a conflict in the 'debian/changelog' file.
        self.assertEquals(conflicts, 1)
        conflict_paths = sorted([c.path for c in ubuntup.conflicts()])
        self.assertEquals(conflict_paths, [u'debian/changelog'])

        # Undo the failed merge.
        ubuntup.revert()

        # The conflict is *not* resolved by calling fix_ancestry_as_needed().
        upstreams_diverged, t_upstream_reverted = MP.fix_ancestry_as_needed(ubuntup, debianp.branch)

        # The ancestry did *not* diverge.
        self.assertEquals(upstreams_diverged, False)
        # The upstreams have not diverged, hence no need to fix/revert 
        # either of them.
        self.assertEquals(t_upstream_reverted, False)

        # Try merging again.
        conflicts = ubuntup.merge_from_branch(
            debianp.branch, to_revision=self.revid_debianp_C)

        # The packaging branch conflict we saw above is still there.
        self.assertEquals(conflicts, 1)
        conflict_paths = sorted([c.path for c in ubuntup.conflicts()])
        self.assertEquals(conflict_paths, [u'debian/changelog'])

    def _setup_debian_upstream_newer(self):
        """
        Set up the following test configuration (debian upstream newer).

        debian-upstream                 ,------------------H
                           A-----------B                    \
        ubuntu-upstream     \           \`-------G           \
                             \           \        \           \
        debian-packaging      \ ,---------D--------\-----------J
                               C                    \
        ubuntu-packaging        `----E---------------I

        where:
             - A = 1.0
             - B = 1.1
             - H = 1.10

             - G = 1.2

             - C = 1.0-1
             - D = 1.1-1
             - J = 1.10-1

             - E = 1.0-1ubuntu1
             - I = 1.2-0ubuntu1

        Please note that the debian and ubuntu *upstream* branches will
        have a conflict with respect to the file 'c'.
        """
        # Set up the debian upstream branch.
        name = 'debu-n'
        vdata = [
            ('upstream-1.0', ('a',), None, None),
            ('upstream-1.1', ('b',), None, None),
            ('upstream-1.10', ('c',), None, None),
            ]
        debu_n = self._setup_branch(name, vdata)

        # Set up the debian packaging branch.
        name = 'debp-n'
        debp_n = self.make_branch_and_tree(name)
        debp_n.pull(debu_n.branch, stop_revision=self.revid_debu_n_A)

        vdata = [
            ('1.0-1', ('debian/', 'debian/changelog'), None, None),
            ('1.1-1', ('o',), debu_n, self.revid_debu_n_B),
            ('1.10-1', ('p',), debu_n, self.revid_debu_n_C),
            ]
        self._setup_branch(name, vdata, debp_n, 'd')

        # Set up the ubuntu upstream branch.
        name = 'ubuu-o'
        ubuu_o = debu_n.bzrdir.sprout(
            name, revision_id=self.revid_debu_n_B).open_workingtree()

        vdata = [
            ('upstream-1.2', ('c',), None, None),
            ]
        self._setup_branch(name, vdata, ubuu_o)

        # Set up the ubuntu packaging branch.
        name = 'ubup-o'
        ubup_o = debu_n.bzrdir.sprout(
            name, revision_id=self.revid_debu_n_A).open_workingtree()

        vdata = [
            ('1.0-1ubuntu1', (), debp_n, self.revid_debp_n_A),
            ('1.2-0ubuntu1', (), ubuu_o, self.revid_ubuu_o_A),
            ]
        self._setup_branch(name, vdata, ubup_o, 'u')

        # Return the ubuntu and the debian packaging branches.
        return (ubup_o, debp_n, ubuu_o, debu_n)

    def _setup_debian_upstream_conflicts(self):
        """
        Set up the following test configuration (debian upstream newer).

        debian-upstream                 ,------------------H
                           A-----------B                    \
        ubuntu-upstream     \           \`-------G           \
                             \           \        \           \
        debian-packaging      \ ,---------D--------\-----------J
                               C                    \
        ubuntu-packaging        `----E---------------I

        where:
             - A = 1.0
             - B = 1.1
             - H = 1.10

             - G = 1.2

             - C = 1.0-1
             - D = 1.1-1
             - J = 1.10-1

             - E = 1.0-1ubuntu1
             - I = 1.2-0ubuntu1

        Please note that the debian upstream and the ubuntu packaging
        branches will have a conflict with respect to the file 'c'.
        """
        # Set up the debian upstream branch.
        name = 'debu-n'
        vdata = [
            ('upstream-1.0', ('a',), None, None),
            ('upstream-1.1', ('b',), None, None),
            ('upstream-1.10', ('c',), None, None),
            ]
        debu_n = self._setup_branch(name, vdata)

        # Set up the debian packaging branch.
        name = 'debp-n'
        debp_n = self.make_branch_and_tree(name)
        debp_n.pull(debu_n.branch, stop_revision=self.revid_debu_n_A)

        vdata = [
            ('1.0-1', ('debian/', 'debian/changelog'), None, None),
            ('1.1-1', ('o',), debu_n, self.revid_debu_n_B),
            ('1.10-1', ('p',), debu_n, self.revid_debu_n_C),
            ]
        self._setup_branch(name, vdata, debp_n, 'd')

        # Set up the ubuntu upstream branch.
        name = 'ubuu-o'
        ubuu_o = debu_n.bzrdir.sprout(
            name, revision_id=self.revid_debu_n_B).open_workingtree()

        vdata = [
            ('upstream-1.2', (), None, None),
            ]
        self._setup_branch(name, vdata, ubuu_o)

        # Set up the ubuntu packaging branch.
        name = 'ubup-o'
        ubup_o = debu_n.bzrdir.sprout(
            name, revision_id=self.revid_debu_n_A).open_workingtree()

        vdata = [
            ('1.0-1ubuntu1', (), debp_n, self.revid_debp_n_A),
            ('1.2-0ubuntu1', ('c',), ubuu_o, self.revid_ubuu_o_A),
            ]
        self._setup_branch(name, vdata, ubup_o, 'u')

        # Return the ubuntu and the debian packaging branches.
        return (ubup_o, debp_n, ubuu_o, debu_n)

    def _setup_debian_upstream_older(self):
        """
        Set up the following test configuration (debian upstream older).

        debian-upstream                 ,----H-------------.
                           A-----------B                    \
        ubuntu-upstream     \           \`-----------G       \
                             \           \            \       \
        debian-packaging      \ ,---------D------------\-------J
                               C                        \
        ubuntu-packaging        `----E-------------------I

        where:
             - A = 1.0
             - B = 1.1
             - H = 1.1.3

             - G = 2.1

             - C = 1.0-1
             - D = 1.1-1
             - J = 1.1.3-1

             - E = 1.0-1ubuntu1
             - I = 2.1-0ubuntu1

        Please note that the debian and ubuntu branches will have a conflict
        with respect to the file 'c'.
        """
        # Set up the debian upstream branch.
        name = 'debu-o'
        vdata = [
            ('upstream-1.0', ('a',), None, None),
            ('upstream-1.1', ('b',), None, None),
            ('upstream-1.1.3', ('c',), None, None),
            ]
        debu_o = self._setup_branch(name, vdata)

        # Set up the debian packaging branch.
        name = 'debp-o'
        debp_o = self.make_branch_and_tree(name)
        debp_o.pull(debu_o.branch, stop_revision=self.revid_debu_o_A)

        vdata = [
            ('1.0-1', ('debian/', 'debian/changelog'), None, None),
            ('1.1-1', ('o',), debu_o, self.revid_debu_o_B),
            ('1.1.3-1', ('p',), debu_o, self.revid_debu_o_C),
            ]
        self._setup_branch(name, vdata, debp_o, 'd')

        # Set up the ubuntu upstream branch.
        name = 'ubuu-n'
        ubuu_n = debu_o.bzrdir.sprout(
            name, revision_id=self.revid_debu_o_B).open_workingtree()

        vdata = [
            ('upstream-2.1', ('c',), None, None),
            ]
        self._setup_branch(name, vdata, ubuu_n)

        # Set up the ubuntu packaging branch.
        name = 'ubup-n'
        ubup_n = debu_o.bzrdir.sprout(
            name, revision_id=self.revid_debu_o_A).open_workingtree()

        vdata = [
            ('1.0-1ubuntu1', (), debp_o, self.revid_debp_o_A),
            ('2.1-0ubuntu1', (), ubuu_n, self.revid_ubuu_n_A),
            ]
        self._setup_branch(name, vdata, ubup_n, 'u')

        # Return the ubuntu and the debian packaging branches.
        return (ubup_n, debp_o, ubuu_n, debu_o)

    def _setup_upstreams_not_diverged(self):
        """
        Set up a test configuration where the usptreams have not diverged.

        debian-upstream                       .-----G
                           A-----------B-----H       \
        ubuntu-upstream     \           \     \       \
                             \           \     \       \
        debian-packaging      \ ,---------D-----\-------J
                               C                 \ 
        ubuntu-packaging        `----E------------I

        where:
             - A = 1.0
             - B = 1.1
             - H = 1.4

             - G = 2.2

             - C = 1.0-1
             - D = 1.1-1
             - J = 2.2-1

             - E = 1.0-1ubuntu1
             - I = 1.4-0ubuntu1

        Please note that there's only one shared upstream branch in this case.
        """
        # Set up the upstream branch.
        name = 'upstream'
        vdata = [
            ('upstream-1.0', ('a',), None, None),
            ('upstream-1.1', ('b',), None, None),
            ('upstream-1.4', ('c',), None, None),
            ]
        upstream = self._setup_branch(name, vdata)

        # Set up the debian upstream branch.
        name = 'dupstream'
        dupstream = upstream.bzrdir.sprout(name).open_workingtree()
        vdata = [
            ('upstream-2.2', (), None, None),
            ]
        dupstream = self._setup_branch(name, vdata, dupstream)

        # Set up the debian packaging branch.
        name = 'debianp'
        debianp = self.make_branch_and_tree(name)
        debianp.pull(dupstream.branch, stop_revision=self.revid_upstream_A)

        vdata = [
            ('1.0-1', ('debian/', 'debian/changelog'), None, None),
            ('1.1-1', ('o',), dupstream, self.revid_upstream_B),
            ('2.2-1', ('p',), dupstream, self.revid_dupstream_A),
            ]
        self._setup_branch(name, vdata, debianp, 'd')

        # Set up the ubuntu packaging branch.
        name = 'ubuntup'
        ubuntup = upstream.bzrdir.sprout(
            name, revision_id=self.revid_upstream_A).open_workingtree()

        vdata = [
            ('1.0-1ubuntu1', (), debianp, self.revid_debianp_A),
            ('1.4-0ubuntu1', (), upstream, self.revid_upstream_C),
            ]
        self._setup_branch(name, vdata, ubuntup, 'u')

        # Return the ubuntu and the debian packaging branches.
        return (ubuntup, debianp)

    def _setup_branch(self, name, vdata, tree=None, log_format=None):
        vids = list(string.ascii_uppercase)
        days = range(len(string.ascii_uppercase))

        if tree is None:
            tree = self.make_branch_and_tree(name)

        tree.lock_write()
        self.addCleanup(tree.unlock)

        def revid_name(vid):
            return 'revid_%s_%s' % (name.replace('-', '_'), vid)

        def add_paths(paths):
            qpaths = ['%s/%s' % (name, path) for path in paths]
            self.build_tree(qpaths)
            tree.add(paths)

        def changelog(vdata, vid):
            result = ''
            day = days.pop(0)
            if isinstance(vdata, tuple):
                uver, dver = vdata[:2]
                ucle = _Ubuntu_changelog % (uver, vid, day)
                dcle = _Debian_changelog % (dver, vid, day)
                result = ucle + dcle
            else:
                if log_format == 'u':
                    result = _Ubuntu_changelog % (vdata, vid, day)
                elif log_format == 'd':
                    result = _Debian_changelog % (vdata, vid, day)

            return result

        def commit(msg, version):
            vid = vids.pop(0)
            if log_format is not None:
                cle = changelog(version, vid)
                p = '%s/work/%s/debian/changelog' % (self.test_base_dir, name)
                _prepend_log(cle, p)
            revid = tree.commit('%s: %s' % (vid, msg))
            setattr(self, revid_name(vid), revid)
            tree.branch.tags.set_tag(version, revid)

        def tree_nick(tree):
            return str(tree)[1:-1].split('/')[-1]

        for version, paths, utree, urevid in vdata:
            msg = ''
            if utree is not None:
                tree.merge_from_branch(utree.branch, to_revision=urevid)
                utree.branch.tags.merge_to(tree.branch.tags)
                if urevid is not None:
                    msg += 'Merged tree %s|%s. ' % (tree_nick(utree), urevid)
                else:
                    msg += 'Merged tree %s. ' % utree
            if paths is not None:
                add_paths(paths)
                msg += 'Added paths: %s. ' % str(paths)

            commit(msg, version)

        return tree


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(MergePackageTests)
    unittest.TextTestRunner(verbosity=2).run(suite)



import os
import sys
import subprocess

from bzrlib import (
    bzrdir,
    revision as mod_revision,
    trace,
    transport,
    workingtree,
    )
from bzrlib import errors as bzr_errors

from bzrlib.plugins.builddeb import (
    default_orig_dir,
    errors,
    import_dsc,
    upstream,
    util,
    )


def _get_tree(package_name):
    try:
        tree = workingtree.WorkingTree.open(".")
    except bzr_errors.NotBranchError:
        if os.path.exists(package_name):
            raise bzr_errors.BzrCommandError("Either run the command from an "
                    "existing branch of upstream, or move %s aside "
                    "and a new branch will be created there."
                    % package_name)
        to_transport = transport.get_transport(package_name)
        tree = to_transport.ensure_base()
        try:
            a_bzrdir = bzrdir.BzrDir.open_from_transport(to_transport)
        except bzr_errors.NotBranchError:
            # really a NotBzrDir error...
            create_branch = bzrdir.BzrDir.create_branch_convenience
            branch = create_branch(to_transport.base,
                                   possible_transports=[to_transport])
            a_bzrdir = branch.bzrdir
        else:
            if a_bzrdir.has_branch():
                raise bzr_errors.AlreadyBranchError(package_name)
            branch = a_bzrdir.create_branch()
            a_bzrdir.create_workingtree()
        try:
            tree = a_bzrdir.open_workingtree()
        except bzr_errors.NoWorkingTree:
            tree = a_bzrdir.create_workingtree()
    return tree


def _get_tarballs(tree, tarball, package_name, version):
    from bzrlib.plugins.builddeb.repack_tarball import repack_tarball
    config = util.debuild_config(tree, tree)
    orig_dir = config.orig_dir or default_orig_dir
    orig_dir = os.path.join(tree.basedir, orig_dir)
    if not os.path.exists(orig_dir):
        os.makedirs(orig_dir)
    format = None
    if tarball.endswith(".tar.bz2") or tarball.endswith(".tbz2"):
        format = "bz2"
    elif tarball.endswith(".tar.xz"):
        format = "xz"
    dest_name = util.tarball_name(package_name, version, None, format=format)
    trace.note("Fetching tarball")
    repack_tarball(tarball, dest_name, target_dir=orig_dir)
    provider = upstream.UpstreamProvider(package_name, version,
            orig_dir, [])
    orig_files = provider.provide(os.path.join(tree.basedir, ".."))
    ret = []
    for filename, component in orig_files:
        ret.append((filename, component, util.md5sum_filename(filename)))
    return ret


def import_upstream(tarball, package_name, version):
    tree = _get_tree(package_name)
    if tree.branch.last_revision() != mod_revision.NULL_REVISION:
        parents = { None: [tree.branch.last_revision()] }
    else:
        parents = {}
    tarball_filenames  = _get_tarballs(tree, tarball,
            package_name, version)
    db = import_dsc.DistributionBranch(tree.branch, tree.branch, tree=tree,
            pristine_upstream_tree=tree)
    dbs = import_dsc.DistributionBranchSet()
    dbs.add_branch(db)
    db.import_upstream_tarballs(tarball_filenames, package_name, version,
        parents)
    return tree


def run_dh_make(tree, package_name, version):
    if not tree.has_filename("debian"):
        tree.mkdir("debian")
    # FIXME: give a nice error on 'debian is not a directory'
    if tree.path2id("debian") is None:
        tree.add("debian")
    command = ["dh_make", "--addmissing", "--packagename",
                "%s_%s" % (package_name, version)]
    if getattr(sys.stdin, 'fileno', None) is None:
        # running in a test or something
        stdin = subprocess.PIPE
        input = "s\n\n"
    else:
        stdin = sys.stdin
        input = None
    try:
        proc = subprocess.Popen(command, cwd=tree.basedir,
                preexec_fn=util.subprocess_setup, stdin=stdin)
    except OSError:
        raise bzr_errors.BzrCommandError("The dh_make command was not found. "
                                         "Please install the dh-make package "
                                         "or use the '--bzr-only' flag and "
                                         "create the debian/ manually.")
    if input is not None:
        proc.stdin.write(input)
        proc.stdin.close()
    retcode = proc.wait()
    if retcode != 0:
        raise bzr_errors.BzrCommandError("dh_make failed.")
    for fn in os.listdir(tree.abspath("debian")):
        if not fn.endswith(".ex") and not fn.endswith(".EX"):
            tree.add(os.path.join("debian", fn))
